'''
Functions for use in pySATSI.

Author: Robert J. Skoumal (rskoumal@usgs.gov)
'''
# External libraries
import numpy as np
import pandas as pd

def az_diff_360(az1,az2):
	'''
	Given two azimuths (0-360), calculates the azimuthal difference between them.
	'''
	_az1=np.min([az1,az2],axis=0)
	_az2=np.max([az1,az2],axis=0)
	tmp_az1=_az2-_az1
	tmp_az2=np.abs(_az2-(_az1+360))
	return np.min([tmp_az1,tmp_az2],axis=0)

def az_diff_180(az1,az2,signed=False):
	'''
	Calculates the difference between two azimuths considering 0/180 edges.
	If signed==True, the sign between az2 and az1 will be preserved with + values = clockwise rotation
	'''
	az1=az1 % 180
	az2=az2 % 180
	if signed==False:
		diff_1=180-np.abs(np.abs(az1-az2)-180)
		diff_2=180-np.abs(np.abs(az1+180-az2)-180)
		difference=np.min((diff_1,diff_2),axis=0)
	else:
		difference = (az2-az1)%180
		# Convert to clockwise difference
		if difference>90:
			difference-=180
	return difference

def aux_plane_ddir(ddir, dip, rake):
	'''
	Converts the aux plane considering the dip direction, dip, and rake angles.

	Args:
		ddir: Dip direction in degrees (float).
		dip: Dip angle in degrees (float).
		rake: Rake angle in degrees (float).

	Returns:
		ddir2: Output array for dip direction
		dip2: Output array for dip angle
		rake2: Output array for rake angle
	'''

	rake[rake==-180]=180

	z=np.radians(ddir)
	dip[dip==90]=89.99999
	z2=np.radians(dip)
	z3=np.radians(rake)

	# Slick vector in plane 1
	s1=-np.cos(z3)*np.cos(z)-np.sin(z3)*np.sin(z)*np.cos(z2)
	s2=np.cos(z3)*np.sin(z)-np.sin(z3)*np.cos(z)*np.cos(z2)
	s3=np.sin(z3)*np.sin(z2)

	# Normal vector to plane 1
	n1=np.sin(z)*np.sin(z2)
	n2=np.cos(z)*np.sin(z2)
	n3=np.cos(z2)

	# Strike vector of plane 2
	h1=-s2
	h2=s1

	z,z2=find_strike_dip(s2,s1,s3)

	z+=90
	ddir2=z
	ddir2[ddir2<0]+=360
	ddir2[ddir2>=360]-=360
	dip2=z2

	z=h1*n1+h2*n2
	z/=np.sqrt(h1**2+h2**2)
	z=np.arccos(z)
	rake2=z
	rake2[s3<0]*=-1
	rake2=np.degrees(rake2)

	tmp_ind=np.where(dip2<0)[0]
	dip2[tmp_ind]*=-1
	ddir2[tmp_ind]+=180
	rake2[tmp_ind]*=-1

	tmp_ind=np.where(dip2>90)[0]
	dip2[tmp_ind]=180-dip2[tmp_ind]
	ddir2[tmp_ind]+=180
	rake2[tmp_ind]*=-1

	ddir2[ddir2>360]-=360
	ddir2[ddir2<0]+=360

	rake2[rake2>360]-=360
	rake2[rake2<0]+=360

	return ddir2,dip2,rake2

def find_strike_dip(n_in, e_in, u_in):
	'''
	Calculates strike and dip angles from a plane's normal vector.

	Args:
		n: North component of normal vector (float).
		e: East component of normal vector (float).
		u: Up component of normal vector (float).

	Returns:
		(strike, dip) in degrees (north of east, dip).
	'''

	n=n_in.copy()
	e=e_in.copy()
	u=u_in.copy()

	# Ensure positive normal vector components
	negu_flag=u<0
	n[negu_flag]*=-1
	e[negu_flag]*=-1
	u[negu_flag]*=-1

	# Strike angle (north of east)
	strike=np.degrees(np.arctan2(e, n))-90

	# Wrap strike angle to 0-360 degrees
	strike%=360

	# Dip angle
	horizontal_magnitude=np.sqrt(n**2 + e**2)
	dip=np.degrees(np.arctan2(horizontal_magnitude, u))

	return strike, dip

def tp2xyz(tr,pl,r):
	'''
	Converts trend (tr) and plunge (pl) to xyz.
	'''
	x=r*np.cos(pl)*np.cos(tr)
	y=r*np.cos(pl)*np.sin(tr)
	z=r*np.sin(pl)
	return x,y,z

def avg_von_mises(azimuth):
	'''
	Computes avg azimuth using von Mises method.
	'''
	rad_azimuth=np.radians(azimuth)
	avg_azimuth=np.degrees(np.arctan(np.sum(np.sin(rad_azimuth))/np.sum(np.cos(rad_azimuth))))
	return avg_azimuth

def calc_shmax_uncert(boot_stress_df,stress_df,p_dict):
	'''
	Calculates SHmax given a dataframe containing the azimuth and plunge
	for S1,S2,S3 and the corresponding min/max values.
	'''
	# If no bootstrapping was done, no uncertainties need to be calculated
	if boot_stress_df['boot'].max()==0:
		shape_ratio=1-stress_df['phi'].values.astype(float)
		stress_df['shmax']=calc_shmax(stress_df[['az1','az2','az3']].values,
								stress_df[['pl1','pl2','pl3']].values,
								shape_ratio)
	else: # Calculates uncertainties
		shape_ratio=1-boot_stress_df['phi'].values.astype(float)
		boot_stress_df['shmax']=calc_shmax(boot_stress_df[['az1','az2','az3']].values,
								boot_stress_df[['pl1','pl2','pl3']].values,
								shape_ratio)
		group_boot_stress_df=boot_stress_df.groupby(by='cluster_id')

		for cluster_x in range(stress_df.shape[0]):
			# Computes avg Shmax
			shmax_azimuth=group_boot_stress_df.get_group(stress_df.loc[cluster_x,'cluster_id'])['shmax'].values
			avg_shmax_azimuth=avg_von_mises(shmax_azimuth)
			stress_df.loc[cluster_x,'shmax']=avg_shmax_azimuth

			# Determine the uncertainty
			n_measurements=shmax_azimuth.shape[0]
			j=int(np.floor((n_measurements-1)*((100.-p_dict['confidence'])/100.)))
			diff_mean=az_diff_180(avg_shmax_azimuth,shmax_azimuth)
			shmax_uncert=diff_mean[np.argsort(np.abs(diff_mean))[::-1][j]]
			stress_df.loc[cluster_x,'shmax_uncert']=shmax_uncert

	# Rounds shmax azimuths
	stress_df['shmax']=np.round(stress_df['shmax'],p_dict['angle_precision'])
	if 'shmax_uncert' in stress_df.columns:
		stress_df['shmax_uncert']=np.round(stress_df['shmax_uncert'],p_dict['angle_precision'])

	return stress_df

def calc_shmax(tr,pl,phi):
	'''
	Given trends and plunges of S1, S2, and S3, calculates SHmax azimuths.
	Input:
		tr: trends, array of [n,3]
		pl: plunges, array of [n,3]
		phi: phi, array of [n]
	output:
		tr_shmax: trend of shmax, array of [n]
	'''
	R=1-phi

	# Converts [trend,plunge] to [x,y].
	x,y,_=tp2xyz(np.radians(tr),np.radians(pl),np.ones(tr.shape))

	# Calculates SHmax following Lund & Townend (2007)
	Y=2.0*(x[:,0]*y[:,0]+(1.0-R)*(x[:,1]*y[:,1]))
	X=x[:,0]**2-y[:,0]**2+(1.0-R)*(x[:,1]**2-y[:,1]**2)
	zero_flag=np.abs(X)<1e-8

	# Handle zero denominator case
	alpha=np.where(zero_flag,np.pi/4.0,np.arctan(Y/X)/2.0)
	special_case_1=np.logical_and(zero_flag,np.abs(x[:,1]*y[:,1])<1e-8)
	alpha=np.where(special_case_1&(np.abs(x[:,0]*y[:,0])<1e-8),-999.,alpha)

	# Handle condition where R is close to 1 + s1Ns1E/s2Ns2E
	if ((x[:,1]==0).any()):
		x[x[:,1]==0,1]=1e-8
	if ((y[:,1]==0).any()):
		y[y[:,1]==0,1]=1e-8
	cond_A=np.abs(R-(1.0+x[:,0]*y[:,0]/(x[:,1]*y[:,1])))<1e-8
	alpha=np.where(cond_A&zero_flag,-999.,alpha)

	# Second derivative and min check
	dev2=-2.0*X*np.cos(2.0*alpha)-2.0*Y*np.sin(2.0*alpha)
	alpha=np.where(dev2>0,alpha+np.pi/2.0,alpha)

	# Ensure alpha is between 0 and 2*pi
	alpha=alpha%(2*np.pi)

	# Converts back to deg
	tr_shmax=np.rad2deg(alpha)

	return tr_shmax

def calc_Aphi(stress_df,p_dict):
	'''
	Uses the orientations of the principal stresses and Phi to quantify the Anderson faulting types
	'''
	# Faulting type characterization following Table 3 from Zoback (1992)
	fault_type_n=np.zeros(stress_df.shape[0]);fault_type_n[:]=np.nan
	fault_type_n[(stress_df['pl1']>=52) & (stress_df['pl3']<=35)]=0
	fault_type_n[(stress_df['pl1']>=40) & (stress_df['pl1']<52) & (stress_df['pl3']<=20)]=0
	fault_type_n[(stress_df['pl1']<40) & (stress_df['pl2']>=45) & (stress_df['pl3']<=20)]=1
	fault_type_n[(stress_df['pl1']<=20) & (stress_df['pl2']>=45) & (stress_df['pl3']<40)]=1
	fault_type_n[(stress_df['pl1']<=20) & (stress_df['pl3']>=40) & (stress_df['pl3']<52)]=2
	fault_type_n[(stress_df['pl1']<=35) & (stress_df['pl3']>=52)]=2

	if np.any(np.isnan(fault_type_n)):
		print('\t*Warning: unable to determine the faulting type needed to compute AΦ for {} cluster(s) using the Zoback (1992) fault regime classification.'.format(np.sum(np.isnan(fault_type_n))))

	# Calculate A_phi following Eq 2 from Simpson (1997)
	Aphi=(fault_type_n+0.5)+((-1)**fault_type_n)*(stress_df['phi']-0.5)
	stress_df.insert(loc=1,column='Aphi',value=Aphi)
	stress_df['Aphi']=stress_df['Aphi'].round(p_dict['vector_precision'])
	stress_df.loc[stress_df['Aphi']<0,'Aphi']=np.nan

	return stress_df


def convert_cluster_ind(mech_df,cluster_df,damp_df,p_dict):
	'''
	Checks p_dict['dimension_names'] and then converts it to [cluster_id]
	'''
	if len(damp_df):
		# If cluster_id is in the mech dataframe, it is required in the damp dataframe too
		if ('cluster_id' in mech_df.columns)!=(('cluster_id1' in damp_df.columns) & ('cluster_id2' in damp_df.columns)):
			raise ValueError(('When using cluster_ids, expecting \'cluster_id\' in the mech file and '+\
					 			'\'cluster_id1\' and \'cluster_id2\' in the damp file.'))

	# If no cluster info is available, then put all mechs in one cluster
	if not(mech_df.columns.isin(p_dict['dimension_names']).any()):
		mech_df['cluster_id']=0
		return mech_df,cluster_df,damp_df

	if not('cluster_id' in mech_df.columns):
		# Ensures the mech and damp dataframes have the same dimension columns
		dim_names=np.asarray(p_dict['dimension_names'])
		mech_dim_names=dim_names[np.isin(dim_names,mech_df.columns)]

		# Converts the dim columns to [cluster_id]
		mech_df=mech_df.sort_values(by=list(mech_dim_names)).reset_index(drop=True)
		cluster_df=mech_df[mech_dim_names].drop_duplicates().reset_index(drop=True)
		cluster_df['cluster_id']=np.arange(len(cluster_df))
		mech_df=mech_df.merge(cluster_df,on=list(mech_dim_names),how='left')

		# Ensures cluster_id is the first column
		cluster_df.insert(0, 'cluster_id', cluster_df.pop('cluster_id'))
		mech_df.insert(0, 'cluster_id', mech_df.pop('cluster_id'))

		if len(damp_df):
			damp_dim_names1=dim_names[np.isin(np.char.add(dim_names,'1'),damp_df.columns)]
			damp_dim_names2=dim_names[np.isin(np.char.add(dim_names,'2'),damp_df.columns)]
			if not(np.array_equal(damp_dim_names1,damp_dim_names2)):
				raise ValueError('Damp file contains unmatching dimension column names.\n\t1: {}\n\t2: {}'.format(damp_dim_names1,damp_dim_names2))
			if not(np.array_equal(mech_dim_names,damp_dim_names1)):
				raise ValueError('Mech and damp files contains columns unmatching dimension names.\n\tMechs: {}\n\tDamp: {}'.format(mech_dim_names,damp_dim_names1))

			tmp_names1=list(np.char.add(mech_dim_names,'1'))
			damp_df=damp_df.merge(cluster_df.add_suffix('1'),on=tmp_names1,how='left')
			damp_df=damp_df.drop(columns=tmp_names1)

			tmp_names2=list(np.char.add(mech_dim_names,'2'))
			damp_df=damp_df.merge(cluster_df.add_suffix('2'),on=tmp_names2,how='left')
			damp_df=damp_df.drop(columns=tmp_names2)

			# Ignore any damping relationships for bins with no mechs
			damp_df=damp_df.drop(damp_df[damp_df.isnull().any(axis=1)].index).reset_index(drop=True).astype(int)
			# Sorts r values for each row
			damp_df[:]=np.sort(damp_df.values,axis=1)
	else:
		cluster_df=pd.DataFrame(columns=['cluster_id'])

	return mech_df,cluster_df,damp_df

def merge_stress_cluster_info(stress_df,cluster_df,p_dict):
	'''
	Combines information from the stress and cluster dataframes together
	'''
	if len(cluster_df):
		merge_col_names=cluster_df.columns[
							cluster_df.columns.isin(['cluster_id','mech_count']) |
							cluster_df.columns.isin(p_dict['dimension_names']) |
							cluster_df.columns.str.startswith('lon') |
							cluster_df.columns.str.startswith('lat') |
							cluster_df.columns.str.startswith('time')
						]
		stress_df=stress_df.merge(cluster_df[merge_col_names],on='cluster_id')
		stress_df.insert(0, 'cluster_id', stress_df.pop('cluster_id'))

	return stress_df


'''
Functions used for computing mechanism RMS.

Author: Robert J. Skoumal (rskoumal@usgs.gov)
'''

# External libraries
import numpy as np

def mech_rms(mech_df,cluster_df,p_dict):
	'''
	Computes the mechanism RMS for each cluster
	'''
	cluster_df['mech_rms']=0.

	# Converts strike,dip,rake to radians
	strike_r=np.deg2rad(mech_df['ddir'].values-90)
	dip_r=np.deg2rad(mech_df['dip'].values)
	rake_r=np.deg2rad(mech_df['rake'].values)

	# Gets fault norm and slip vector
	fnorm,slip=vector_from_sdr(strike_r,dip_r,rake_r)

	if not('mech_count' in cluster_df.columns):
		cluster_df=cluster_df.merge(mech_df['cluster_id'].value_counts().rename('mech_count'),on='cluster_id',how='left')

	# Determines which clusters have mechanisms
	consider_cluster_ind=cluster_df[cluster_df['mech_count']>0].index.values

	# Loops over each cluster with mechs, considering the mechanisms from each
	group_mech_df=mech_df.groupby(by='cluster_id')
	for cluster_x in consider_cluster_ind:
		tmp_mech_df=group_mech_df.get_group(cluster_df.loc[cluster_x,'cluster_id'])
		mech_index=tmp_mech_df.index.values

		# Computes the average mechanism
		norm1_avg,norm2_avg=average_mech(fnorm[:,mech_index],slip[:,mech_index])

		# Finds the min angle between the avg mech and mech population
		rota,norm2,slip2=mech_rotation(norm1_avg,fnorm[:,mech_index],norm2_avg,slip[:,mech_index])

		# Computes mechanism RMS
		cluster_df.loc[cluster_x,'mech_rms']=np.sqrt(np.mean(rota**2))

	cluster_df['mech_rms']=cluster_df['mech_rms'].round(p_dict['angle_precision'])

	return cluster_df

def vector_from_sdr(strike_r,dip_r,rake_r):
	'''
	Gets fault normal vector (fnorm) and slip vector (slip) from [strike,dip,rake] (in radians).
	Uses (x,y,z) coordinate system with x=north, y=east, z=down
		Reference:  Aki and Richards, p. 115
	Based on code from HASH (Hardebeck & Shearer, 2002) and SKHASH (Skoumal et al., 2024)
	'''
	fnorm=np.zeros((3,len(strike_r)))
	slip=np.zeros((3,len(strike_r)))

	fnorm[0,:]=-np.sin(dip_r)*np.sin(strike_r)
	fnorm[1,:]=np.sin(dip_r)*np.cos(strike_r)
	fnorm[2,:]=-np.cos(dip_r)
	slip[0,:]=np.cos(rake_r)*np.cos(strike_r)+np.cos(dip_r)*np.sin(rake_r)*np.sin(strike_r)
	slip[1,:]=np.cos(rake_r)*np.sin(strike_r)-np.cos(dip_r)*np.sin(rake_r)*np.cos(strike_r)
	slip[2,:]=-np.sin(rake_r)*np.sin(dip_r)

	return fnorm,slip

def average_mech(norm1in,norm2in):
	'''
	Computes the average mech of the solutions.

	Inputs:
		norm1in: normal to fault plane, array(3,nf)
		norm2in: slip vector, array(3,nf)
	Output:
		norm1_avg: normal to average of plane 1
		norm2_avg: normal to average of plane 2

	Based on code from SKHASH (Skoumal et al., 2024)
	'''
	if norm1in.shape != norm2in.shape:
		raise ValueError('***Error in average_mech: shape of norm1in and norm2in must be the same')
	if len(norm1in.shape)!=2:
		raise ValueError('***Error in average_mech: norm1in and norm2in must each be an array of shape n-by-3')
	if norm1in.shape[0]!=3:
		raise ValueError('***Error in average_mech: norm1in and norm2in must each be an array of shape n-by-3')

	norm1=norm1in.copy()
	norm2=norm2in.copy()

	# If there is only one mechanism, return that mechanism
	if norm1.shape[1]==1:
		return norm1[:,0],norm2[:,0]

	norm1_ref=norm1in[:,0].copy()
	norm2_ref=norm2in[:,0].copy()

	rota,temp1,temp2=mech_rotation(norm1_ref,norm1[:,1:],norm2_ref,norm2[:,1:])

	norm1_avg=np.sum(np.hstack((norm1[:,[0]],temp1)),axis=1)
	norm2_avg=np.sum(np.hstack((norm2[:,[0]],temp2)),axis=1)
	ln_norm1=np.sqrt(np.sum(norm1_avg**2))
	ln_norm2=np.sqrt(np.sum(norm2_avg**2))
	norm1_avg=norm1_avg/ln_norm1
	norm2_avg=norm2_avg/ln_norm2

	# Determine the RMS observed angular difference between the average
	# Normal vectors and the normal vectors of each mechanism
	rota,temp1,temp2=mech_rotation(norm1_avg,norm1,norm2_avg,norm2)
	d11=temp1[0,:]*norm1_avg[0]+temp1[1,:]*norm1_avg[1]+temp1[2,:]*norm1_avg[2]
	d22=temp2[0,:]*norm2_avg[0]+temp2[1,:]*norm2_avg[1]+temp2[2,:]*norm2_avg[2]

	d11[d11>1]=1
	d11[d11<-1]=-1
	d22[d22>1]=1
	d22[d22<-1]=-1

	a11=np.arccos(d11)
	a22=np.arccos(d22)

	avang1=np.sqrt(np.sum(a11**2)/len(a11))
	avang2=np.sqrt(np.sum(a22**2)/len(a22))

	# the average normal vectors may not be exactly orthogonal (although
	# usually they are very close) - find the misfit from orthogonal and
	# adjust the vectors to make them orthogonal - adjust the more poorly
	# constrained plane more
	if (avang1+avang2)>=0.0001:
		maxmisf=0.01
		fract1=avang1/(avang1+avang2)
		for icount in range(100):
			dot1=norm1_avg[0]*norm2_avg[0]+norm1_avg[1]*norm2_avg[1]+norm1_avg[2]*norm2_avg[2]
			misf=90-np.rad2deg(np.arccos(dot1))
			if abs(misf)<=maxmisf:
				break
			else:
				theta1=np.deg2rad(misf*fract1)
				theta2=np.deg2rad(misf*(1-fract1))
				temp=norm1_avg
				norm1_avg=norm1_avg-norm2_avg*np.sin(theta1)
				norm2_avg=norm2_avg-temp*np.sin(theta2)
				ln_norm1=np.sqrt(np.sum(norm1_avg*norm1_avg))
				ln_norm2=np.sqrt(np.sum(norm2_avg*norm2_avg))
				norm1_avg=norm1_avg/ln_norm1
				norm2_avg=norm2_avg/ln_norm2
	return norm1_avg,norm2_avg

def mech_rotation(norm1_in,norm2_in,slip1_in,slip2_in):
	'''
	Finds the minimum rotation angle between two mechanisms.
	Does not assume that the normal and slip vectors are matched.
	Input:
		norm1_in: normal to fault plane 1
		norm2_in: normal to fault plane 2
		slip1_in: slip vector 1
		slip2_in: slip vector 2
	Output:
		rota: rotation angle
		norm2: normal to fault plane, best combination
		slip2: slip vector, best combination

	Based on code from SKHASH (Skoumal et al., 2024)
	'''
	if norm1_in.shape != slip1_in.shape:
		raise ValueError('***Error in mech_rotation: shape of norm1_in and slip1_in must be the same')
	if norm2_in.shape != slip2_in.shape:
		raise ValueError('***Error in mech_rotation: shape of norm2_in and slip2_in must be the same')
	if len(norm1_in.shape)!=1:
		raise ValueError('***Error in mech_rotation: norm1_in and slip1_in must each be an array of length 3')
	if norm1_in.shape[0]!=3:
		raise ValueError('***Error in mech_rotation: norm1_in and slip1_in must each be an array of length 3')
	if len(norm2_in.shape)!=2:
		raise ValueError('***Error in mech_rotation: norm2_in and slip2_in must each be an array of shape 3-by-n')
	if norm2_in.shape[0]!=3:
		raise ValueError('***Error in mech_rotation: norm2_in and slip2_in must each be an array of shape 3-by-n')

	norm1=norm1_in.copy()
	norm2=norm2_in.copy().T
	slip1=slip1_in.copy()
	slip2=slip2_in.copy().T

	num_vect=norm2.shape[0]
	rotemp=np.zeros((num_vect,4))
	for iter_x in range(0,4): # Iteration over the 4 possibilities
		if iter_x<2:
			norm2_temp=norm2.copy()
			slip2_temp=slip2.copy()
		else:
			norm2_temp=slip2.copy()
			slip2_temp=norm2.copy()
		if (iter_x==1) | (iter_x==3):
			norm2_temp=-norm2_temp
			slip2_temp=-slip2_temp

		B1=np.cross(slip1,norm1)*-1
		B2=np.cross(slip2_temp,norm2_temp)*-1

		phi=np.zeros((num_vect,3))
		phi[:,0]=norm1[0]*norm2_temp[:,0]+norm1[1]*norm2_temp[:,1]+norm1[2]*norm2_temp[:,2]
		phi[:,1]=slip1[0]*slip2_temp[:,0]+slip1[1]*slip2_temp[:,1]+slip1[2]*slip2_temp[:,2]
		phi[:,2]=B1[0]*B2[:,0]+B1[1]*B2[:,1]+B1[2]*B2[:,2]
		phi[phi>1]=1
		phi[phi<-1]=-1
		phi=np.arccos(phi)

		phi_flag=(phi<(1e-3))
		# if the mechanisms are very close, rotation = 0. Otherwise, calculate the rotation
		rot_ind=np.where(np.any(~phi_flag,axis=1))[0]

		# if one vector is the same, it is the rotation axis
		tmp_ind=rot_ind[np.where(phi_flag[rot_ind,2])[0]]
		rotemp[tmp_ind,iter_x]=(phi[tmp_ind,0])
		tmp_ind=rot_ind[np.where(phi_flag[rot_ind,0])[0]]
		rotemp[tmp_ind,iter_x]=(phi[tmp_ind,1])
		tmp_ind=rot_ind[np.where(phi_flag[rot_ind,1])[0]]
		rotemp[tmp_ind,iter_x]=(phi[tmp_ind,2])

		# find difference vectors - the rotation axis must be orthogonal to all three vectors
		rot_ind=np.where(np.all(~phi_flag,axis=1))[0]

		if len(rot_ind)==0:
			continue

		n=np.zeros((len(rot_ind),3,3))
		n[:,:,0]=norm1-norm2_temp[rot_ind,:]
		n[:,:,1]=slip1-slip2_temp[rot_ind,:]
		n[:,:,2]=B1-B2[rot_ind,:]
		scale=np.sqrt(n[:,0,:]**2+n[:,1,:]**2+n[:,2,:]**2)
		n=n/scale[:,np.newaxis,:]

		qdot=np.zeros((len(rot_ind),3))
		qdot[:,2]=n[:,0,0]*n[:,0,1]+n[:,1,0]*n[:,1,1]+n[:,2,0]*n[:,2,1]
		qdot[:,1]=n[:,0,0]*n[:,0,2]+n[:,1,0]*n[:,1,2]+n[:,2,0]*n[:,2,2]
		qdot[:,0]=n[:,0,1]*n[:,0,2]+n[:,1,1]*n[:,1,2]+n[:,2,1]*n[:,2,2]

		# use the two largest difference vectors, as long as they aren't orthogonal
		iout=np.zeros(len(rot_ind),dtype=int)-1
		qdot_flag=np.any(qdot>0.9999,axis=1)
		tmp_row=np.where(qdot_flag)[0]
		if len(tmp_row)>0:
			iout[tmp_row]=np.argmax(qdot[tmp_row,:],axis=1)
		tmp_row=np.where(~qdot_flag)[0]
		if len(tmp_row)>0:
			iout[tmp_row]=np.argmin(scale[tmp_row,:],axis=1)

		n1=np.zeros((len(rot_ind),3))
		n2=np.zeros((len(rot_ind),3))
		k=np.ones(len(rot_ind),dtype=bool)
		for j in range(3):
			tmp_ind=np.where(j!=iout)[0]
			tmp_ind_1=tmp_ind[k[tmp_ind]==True]
			tmp_ind_2=tmp_ind[k[tmp_ind]==False]

			if len(tmp_ind_1)>0:
				n1[tmp_ind_1,:]=n[tmp_ind_1,:,j]
				k[tmp_ind_1]=False
			if len(tmp_ind_2)>0:
				n2[tmp_ind_2,:]=n[tmp_ind_2,:,j]

		#  find rotation axis by taking cross product
		R=np.cross(n2,n1)*-1
		scaleR=np.sqrt(np.sum(R**2,axis=1))

		if np.any(scaleR==0):
			tmp_ind=np.where(scaleR==0)[0]
			rotemp[rot_ind[tmp_ind],iter_x]=9999
			rot_ind=np.delete(rot_ind,tmp_ind)
			scaleR=np.delete(scaleR,tmp_ind)
			R=np.delete(R,tmp_ind,axis=0)

		R=R/scaleR[:,np.newaxis]
		theta=np.zeros((len(rot_ind),3))
		theta[:,0]=norm1[0]*R[:,0]+norm1[1]*R[:,1]+norm1[2]*R[:,2]
		theta[:,1]=slip1[0]*R[:,0]+slip1[1]*R[:,1]+slip1[2]*R[:,2]
		theta[:,2]=B1[0]*R[:,0]+B1[1]*R[:,1]+B1[2]*R[:,2]
		theta[theta>1]=1
		theta[theta<-1]=-1
		theta=np.arccos(theta)

		iuse=np.argmin(np.abs(theta-(np.pi/2)),axis=1)
		tmp_ind=np.arange(len(iuse))

		tmp_rotemp=(np.cos(phi[rot_ind,iuse])-np.cos(theta[tmp_ind,iuse])**2)/(np.sin(theta[tmp_ind,iuse])**2)
		tmp_rotemp[tmp_rotemp>1]=1
		tmp_rotemp[tmp_rotemp<-1]=-1
		tmp_rotemp=np.arccos(tmp_rotemp)
		rotemp[rot_ind,iter_x]=tmp_rotemp

	rotemp=np.rad2deg(rotemp)
	rotemp=np.abs(rotemp)
	irot=np.argmin(rotemp,axis=1)

	tmp_ind=np.arange(len(irot))
	rota=rotemp[tmp_ind,irot]

	tmp_ind=np.where(irot>=2)[0]
	qtemp=slip2[tmp_ind,:]
	slip2[tmp_ind,:]=norm2[tmp_ind,:]
	norm2[tmp_ind,:]=qtemp

	tmp_ind=np.where( (irot==1) | (irot==3) )[0]
	norm2[tmp_ind,:]*=-1
	slip2[tmp_ind,:]*=-1

	return rota,norm2.T,slip2.T



'''
Functions for clustering data.

Author: Robert J. Skoumal (rskoumal@usgs.gov)
'''

# Standard libraries
import itertools
from timeit import default_timer as timer

# External libraries
import numpy as np
import pandas as pd
from sklearn.utils.extmath import weighted_mode
from sklearn.preprocessing import LabelEncoder

# Local libraries
import functions.checkerboard as checkerboard


def grid(mech_df,p_dict):
	'''
	Produces (non)uniform spatial and/or temporal grids.

	Input:
		mech_df: DataFrame with columns: ['cluster_id','strike','dip','rake']
		p_dict: Parameter dictionary
	Output:
		cluster_df: DataFrame with columns ['cluster_id','lon1','lon2','lat1','lat2']
	'''
	if ((not(p_dict['do_temporal_gridding'])) &
		(not(p_dict['do_vertical_gridding'])) &
		(not(p_dict['do_horizontal_gridding'])) &
		(not(mech_df.columns.isin(p_dict['dimension_names']).any()))):

		cluster_df=pd.DataFrame(data={'cluster_id':[0],'mech_count':[len(mech_df)]},dtype=int)
		return mech_df,cluster_df,pd.DataFrame(),p_dict

	flag_deg_to_km=False
	flag_cluster_x=False
	flag_cluster_y=False

	mech_df['cluster_id']=-999
	cluster_col=[]
	if len(p_dict['xspace']):
		flag_cluster_x=True
		mech_df['x']=-1
		cluster_col.append('x')
	if len(p_dict['yspace']):
		flag_cluster_y=True
		mech_df['y']=-1
		cluster_col.append('y')
	if p_dict['do_vertical_gridding']:
		mech_df['z']=-1
		cluster_col.append('z')
	if (len(p_dict['xspace'])>1) | (len(p_dict['yspace'])>1):
		mech_df['l']=-1
		cluster_col.append('l')
	if p_dict['do_temporal_gridding']:
		mech_df['t']=-1
		cluster_col.append('t')

	# Temporally clusters mechanisms
	if p_dict['do_temporal_gridding']:
		num_tbreaks=len(p_dict['time_breaks'])
		if not(pd.api.types.is_datetime64_any_dtype(mech_df['time'])):
			mech_df['time']=pd.to_datetime(mech_df['time'])

		mech_df.loc[mech_df['time']<p_dict['time_breaks'][0],'t']=0
		mech_df.loc[mech_df['time']>=p_dict['time_breaks'][-1],'t']=len(p_dict['time_breaks'])
		for t_x in range(1,len(p_dict['time_breaks'])):
			mech_df.loc[(mech_df['time']>=p_dict['time_breaks'][t_x-1]) &
						(mech_df['time']<p_dict['time_breaks'][t_x]),'t']=t_x

		time_cluster_df=pd.DataFrame(index=np.arange(len(p_dict['time_breaks'])+1),columns=['t','time1','time2'])
		time_cluster_df['time1']=time_cluster_df['time1'].astype('datetime64[ns]')
		time_cluster_df['time2']=time_cluster_df['time2'].astype('datetime64[ns]')
		time_cluster_df['t']=np.arange(time_cluster_df.shape[0])
		time_cluster_df.loc[1:,'time1']=p_dict['time_breaks']
		time_cluster_df.loc[:(time_cluster_df.shape[0]-2),'time2']=p_dict['time_breaks']
		time_cluster_df.loc[0,'time1']=p_dict['earliest_event_time']
		time_cluster_df.loc[time_cluster_df.index[-1],'time2']=p_dict['latest_event_time']

		# Fills empty times with default time values
		time_cluster_df['time1']=time_cluster_df['time1'].fillna(pd.Timestamp('1700-01-01'))
		time_cluster_df['time2']=time_cluster_df['time2'].fillna(pd.Timestamp('2200-01-01'))

	# Vertically clusters mechanisms
	if p_dict['do_vertical_gridding']:
		num_zbreaks=len(p_dict['depth_breaks'])
		mech_df.loc[mech_df['depth']<p_dict['depth_breaks'][0],'z']=0
		mech_df.loc[mech_df['depth']>=p_dict['depth_breaks'][-1],'z']=len(p_dict['depth_breaks'])
		for z_x in range(1,len(p_dict['depth_breaks'])):
			mech_df.loc[(mech_df['depth']>=p_dict['depth_breaks'][z_x-1]) &
						(mech_df['depth']<p_dict['depth_breaks'][z_x]),'t']=z_x

		depth_cluster_df=pd.DataFrame(index=np.arange(len(p_dict['depth_breaks'])+1),columns=['z','depth1','depth2'],dtype='float',data=-999)
		depth_cluster_df['z']=np.arange(depth_cluster_df.shape[0]).astype(int)
		depth_cluster_df.loc[1:,'depth1']=p_dict['depth_breaks']
		depth_cluster_df.loc[:(depth_cluster_df.shape[0]-2),'depth2']=p_dict['depth_breaks']
		depth_cluster_df.loc[0,'depth1']=p_dict['min_depth']
		depth_cluster_df.loc[depth_cluster_df.index[-1],'depth2']=p_dict['max_depth']

	# Horizontally clusters mechanisms
	if p_dict['do_horizontal_gridding']:
		if 'lon' in mech_df.columns:
			if not(np.isfinite(p_dict['min_lon'])):
				p_dict['min_lon']=mech_df['lon'].min()
			if not(np.isfinite(p_dict['max_lon'])):
				p_dict['max_lon']=mech_df['lon'].max()
		if 'lat' in mech_df.columns:
			if not(np.isfinite(p_dict['min_lat'])):
				p_dict['min_lat']=mech_df['lat'].min()
			if not(np.isfinite(p_dict['max_lat'])):
				p_dict['max_lat']=mech_df['lat'].max()
			if not(np.isfinite(p_dict['base_lat'])):
				# If study area is provided, takes the avg lat of the study area
				if np.isfinite([p_dict['min_lat'],p_dict['max_lat']]).all():
					p_dict['base_lat']=np.abs(np.mean([p_dict['min_lat'],p_dict['max_lat']]))
				# Otherwise determine the avg lat of the mechanisms
				elif 'lat' in mech_df.columns:
					p_dict['base_lat']=mech_df['lat'].mean()

		horizontal_tmp_col=[]
		if flag_cluster_x:
			horizontal_tmp_col.append('_xn')
		if flag_cluster_y:
			horizontal_tmp_col.append('_yn')

		# Converts deg to km
		if flag_cluster_x:
			if not('x_km' in mech_df.columns) and ('lon' in mech_df.columns):
				lonmult=111.1*np.cos(np.pi*p_dict['base_lat']/180.0)
				mech_df['x_km']=(mech_df['lon']-p_dict['min_lon'])*lonmult
				flag_deg_to_km=True
		if flag_cluster_y:
			if not('y_km' in mech_df.columns) and ('lat' in mech_df.columns):
				latmult=111.1
				mech_df['y_km']=(mech_df['lat']-p_dict['min_lat'])*latmult
				flag_deg_to_km=True

		if not(p_dict['do_temporal_gridding']):
			num_tbreaks=0
		if not(p_dict['do_vertical_gridding']):
			num_zbreaks=0

		# Rotates points
		if p_dict['rotate_deg']!=0:
			mech_df[['x_km','y_km']]=rotate_points([0,0],mech_df[['x_km','y_km']].values,p_dict['rotate_deg'])

		num_hspace=np.max([len(p_dict['xspace']),len(p_dict['yspace'])])
		if num_hspace==1:
			print('\tHorizontal spacing: {} km'.format(np.max([p_dict['xspace'],p_dict['yspace']])))
		for t_ind in range(num_tbreaks+1):
			for z_ind in range(num_zbreaks+1):
				for h_ind in range(num_hspace):
					print_hspace=0
					if flag_cluster_x:
						xspace=p_dict['xspace'][h_ind]
						print_hspace=xspace
					if flag_cluster_y:
						yspace=p_dict['yspace'][h_ind]
						print_hspace=yspace

					tmp_mech_df=mech_df.loc[(mech_df['cluster_id']<-1)].copy()
					if num_hspace>1:
						print('\tHorizontal {}/{}: [{} km]'.format(h_ind,num_hspace-1,print_hspace))
					if p_dict['do_vertical_gridding']:
						if not(np.isfinite(depth_cluster_df.loc[z_ind,'depth1'])):
							print('\tVertical {}/{}: [<{} km]'.format(z_ind,num_zbreaks,depth_cluster_df.loc[z_ind,'depth2']))
						elif not(np.isfinite(depth_cluster_df.loc[z_ind,'depth2'])):
							print('\tVertical {}/{}: [>{} km]'.format(z_ind,num_zbreaks,depth_cluster_df.loc[z_ind,'depth1']))
						else:
							print('\tVertical {}/{}: [{} - {} km]'.format(z_ind,num_zbreaks,depth_cluster_df.loc[z_ind,'depth1'],depth_cluster_df.loc[z_ind,'depth2']))
						tmp_mech_df=tmp_mech_df.loc[(tmp_mech_df['z']==z_ind)]
					if p_dict['do_temporal_gridding']:
						print('\tTemporal {}/{}: [{} - {}]'.format(t_ind,num_tbreaks,time_cluster_df.loc[t_ind,'time1'],time_cluster_df.loc[t_ind,'time2']))
						tmp_mech_df=tmp_mech_df.loc[(tmp_mech_df['t']==t_ind)]

					# Determines which grid cell the considered events fall into
					if flag_cluster_x:
						tmp_mech_df['_xn']=np.floor((tmp_mech_df['x_km'].values)/(xspace)).astype(int)
					if flag_cluster_y:
						tmp_mech_df['_yn']=np.floor((tmp_mech_df['y_km'].values)/(yspace)).astype(int)

					# Selects only the clusters with >= $minenv
					tmp_mech_df=tmp_mech_df.loc[tmp_mech_df.groupby(horizontal_tmp_col).transform('size').ge(p_dict['minev'])]

					# Records the grid cell each selected earthquake falls into
					if flag_cluster_x:
						mech_df.loc[tmp_mech_df.index,'x']=tmp_mech_df['_xn']
					if flag_cluster_y:
						mech_df.loc[tmp_mech_df.index,'y']=tmp_mech_df['_yn']
					if num_hspace>1:
						mech_df.loc[tmp_mech_df.index,'l']=h_ind

					mech_df.loc[tmp_mech_df.index,'cluster_id']=-1 # Signifies mech is clustered, but no id yet

	rm_index=mech_df[(mech_df.filter(['x','y','z','l','t'])==-1).any(axis=1)].index
	if len(rm_index):
		mech_df=mech_df.drop(rm_index).reset_index(drop=True)
		print('\tDiscarding {}/{} mechanisms ({}%) due to not being in a cluster with at least {} events.'.format(len(rm_index),len(mech_df),np.round(len(rm_index)/len(mech_df)*100,1),p_dict['minev']))

	# Determines cluster ids for each event
	sort_col=list(mech_df.filter(p_dict['dimension_names']).columns)
	mech_df=mech_df.sort_values(sort_col).reset_index(drop=True)
	mech_df['cluster_id']=mech_df.groupby(sort_col).ngroup()

	# Drops any unclustered earthquakes
	tmp_counts_df=mech_df['cluster_id'].value_counts()
	rm_index=tmp_counts_df.loc[tmp_counts_df<p_dict['minev']].index
	if len(rm_index):
		discard_ind=mech_df[mech_df['cluster_id'].isin(rm_index)].index
		print('\tDiscarding {}/{} mechanisms ({}%) due to not being in a cluster with at least {} events.'.format(len(discard_ind),len(mech_df),np.round(len(discard_ind)/len(mech_df)*100,1),p_dict['minev']))
		mech_df=mech_df.drop(discard_ind).reset_index(drop=True)

	if len(mech_df)==0:
		return pd.DataFrame(),pd.DataFrame(),pd.DataFrame(),p_dict

	# Remove unneeded spatial columns
	if flag_deg_to_km:
		mech_df=mech_df.drop(columns=mech_df.filter(['x_km','y_km']))

	# Creates cluster information
	cluster_df=mech_df.filter(np.hstack(['cluster_id',p_dict['dimension_names']])).drop_duplicates(subset=['cluster_id'])
	cluster_df=cluster_df.sort_values(by='cluster_id').reset_index(drop=True)

	if p_dict['do_damping'] and (not(len(p_dict['damp_file']))):
		cluster_df,damp_df=damp_connect(cluster_df,p_dict)
	else:
		damp_df=pd.DataFrame(columns=['cluster_id1','cluster_id2'])
	if len(damp_df)==0:
		if p_dict['do_damping']:
			p_dict['do_damping']=False
			print('\t\tNo damping needed.')

	# Determines the lon coords of the clusters
	if flag_cluster_x:
		if 'x' in cluster_col:
			lonmult=111.1*np.cos(np.pi*p_dict['base_lat']/180.0)
			if 'l' in cluster_col:
				cluster_df['x1']=cluster_df['x']*p_dict['xspace'][cluster_df['l']]
				cluster_df['x2']=(cluster_df['x']+1)*p_dict['xspace'][cluster_df['l']]
			else:
				cluster_df['x1']=cluster_df['x']*p_dict['xspace'][0]
				cluster_df['x2']=(cluster_df['x']+1)*p_dict['xspace'][0]
		elif 'y' in cluster_col:
			cluster_df['lon1']=mech_df['lon'].min()
			cluster_df['lon2']=mech_df['lon'].max()

	# Determines the lat coords of the clusters
	if flag_cluster_y:
		if 'y' in cluster_col:
			latmult=111.1
			if 'l' in cluster_col:
				cluster_df['y1']=cluster_df['y']*p_dict['yspace'][cluster_df['l']]#/latmult+p_dict['min_lat']
				cluster_df['y2']=(cluster_df['y']+1)*p_dict['yspace'][cluster_df['l']]#/latmult+p_dict['min_lat']
			else:
				cluster_df['y1']=cluster_df['y']*p_dict['yspace'][0]
				cluster_df['y2']=(cluster_df['y']+1)*p_dict['yspace'][0]
		elif 'x' in cluster_col:
			cluster_df['lat1']=mech_df['lat'].min()
			cluster_df['lat2']=mech_df['lat'].max()

	# Converts grid x,y coords to lon,lat
	if p_dict['rotate_deg']==0:
		if flag_cluster_x:
			if ('x1' in cluster_df.columns) & ('lon' in mech_df.columns):
				cluster_df[['lon1','lon2']]=cluster_df[['x1','x2']]/lonmult+p_dict['min_lon']
				cluster_df[['lon1','lon2']]=cluster_df[['lon1','lon2']].round(p_dict['coordinate_precision'])
		if flag_cluster_y:
			if ('y1' in cluster_df.columns) & ('lat' in mech_df.columns):
				cluster_df[['lat1','lat2']]=cluster_df[['y1','y2']]/latmult+p_dict['min_lat']
				cluster_df[['lat1','lat2']]=cluster_df[['lat1','lat2']].round(p_dict['coordinate_precision'])
	else:
		# Rotates the grid back
		cluster_df[['x1_r','y1_r']]=rotate_points([0,0],cluster_df[['x1','y1']].values,-p_dict['rotate_deg'])
		cluster_df[['x2_r','y2_r']]=rotate_points([0,0],cluster_df[['x2','y2']].values,-p_dict['rotate_deg'])
		cluster_df[['x3_r','y3_r']]=rotate_points([0,0],cluster_df[['x1','y2']].values,-p_dict['rotate_deg'])
		cluster_df[['x4_r','y4_r']]=rotate_points([0,0],cluster_df[['x2','y1']].values,-p_dict['rotate_deg'])
		cluster_df[['lon1','lon2','lon3','lon4']]=cluster_df[['x1_r','x2_r','x3_r','x4_r']]/lonmult+p_dict['min_lon']
		cluster_df[['lat1','lat2','lat3','lat4']]=cluster_df[['y1_r','y2_r','y3_r','y4_r']]/latmult+p_dict['min_lat']
		cluster_df[['lon1','lat1','lon2','lat2']]=cluster_df[['lon1','lat1','lon2','lat2']].round(p_dict['coordinate_precision'])
		cluster_df[['lon3','lat3','lon4','lat4']]=cluster_df[['lon3','lat3','lon4','lat4']].round(p_dict['coordinate_precision'])

		# Removes unneeded columns
		cluster_df=cluster_df.drop(columns=cluster_df.filter(['x1_r','y1_r','x2_r','y2_r','x3_r','y3_r','x4_r','y4_r']))

	if p_dict['do_vertical_gridding']:
		if 'z' in cluster_col:
			cluster_df=cluster_df.merge(depth_cluster_df,on='z',how='left')

	if p_dict['do_temporal_gridding']:
		if 't' in cluster_col:
			cluster_df=cluster_df.merge(time_cluster_df,on='t',how='left')

	# Ensures all the clustering went appropriately
	if cluster_df.isnull().values.any():
		raise ValueError('Null value(s) present in cluster records.')

	# Counts the number of mechanisms in each cluster
	cluster_df=cluster_df.merge(mech_df['cluster_id'].value_counts().rename('mech_count'),on='cluster_id',how='left')
	cluster_df['mech_count']=cluster_df['mech_count'].fillna(0).astype(int)

	if len(p_dict['checkerboard_type']):
		# Finds the grid interval for the checkerboard
		eq_grid_ind=np.where(p_dict['checkerboard_hspace']==p_dict['xspace'])[0]
		lt_grid_ind=np.where((p_dict['checkerboard_hspace'] % p_dict['xspace'])==0)[0]

		if len(eq_grid_ind):
			grid_ind=eq_grid_ind[0]
		elif len(lt_grid_ind):
			grid_ind=lt_grid_ind[0]
		else:
			raise ValueError('Checkerboard spacing ($checkerboard_hspace={}) must be both >= than and a factor of the horizontal grid spacing ($xspace={})'.format(p_dict['checkerboard_hspace'],p_dict['xspace']))

		if len(eq_grid_ind) | len(lt_grid_ind):
			grid_spacing=int(p_dict['checkerboard_hspace']/p_dict['xspace'][grid_ind])
		else:
			grid_spacing=int(p_dict['xspace'][grid_ind]/p_dict['checkerboard_hspace'])

		min_x=np.min([mech_df['x'].min(),cluster_df['x'].min()])
		min_y=np.min([mech_df['y'].min(),cluster_df['y'].min()])

		# Determines traditional grid Shmax
		if p_dict['checkerboard_type']=='traditional':

			x_grid_flag=(((mech_df['x'].values-min_x) / grid_spacing).astype(int) % 2).astype(bool)
			y_grid_flag=(((mech_df['y'].values-min_y) / grid_spacing).astype(int) % 2).astype(bool)
			grid_flag=(x_grid_flag & ~y_grid_flag) | (~x_grid_flag & y_grid_flag)

			cluster_x_grid_flag=(((cluster_df['x'].values-min_x) / grid_spacing).astype(int) % 2).astype(bool)
			cluster_y_grid_flag=(((cluster_df['y'].values-min_y) / grid_spacing).astype(int) % 2).astype(bool)
			cluster_grid_flag=(cluster_x_grid_flag & ~cluster_y_grid_flag) | (~cluster_x_grid_flag & cluster_y_grid_flag)

			mech_df['synth_shmax']=p_dict['checkerboard_shmax1']
			mech_df.loc[~grid_flag,'synth_shmax']=p_dict['checkerboard_shmax2']

			cluster_df['synth_shmax']=p_dict['checkerboard_shmax1']
			cluster_df.loc[~cluster_grid_flag,'synth_shmax']=p_dict['checkerboard_shmax2']

		# Creates gradient checkerboard
		elif p_dict['checkerboard_type']=='gradient':

			x_grid=((mech_df['x'].values-min_x) / grid_spacing).astype(int)
			y_grid=((mech_df['y'].values-min_y) / grid_spacing).astype(int)
			mech_df['synth_shmax']=p_dict['checkerboard_shmax_start']+p_dict['checkerboard_gradient']*np.max(np.vstack((x_grid,y_grid)),axis=0)

			cluster_x_grid=((cluster_df['x'].values-min_x) / grid_spacing).astype(int)
			cluster_y_grid=((cluster_df['y'].values-min_y) / grid_spacing).astype(int)
			cluster_df['synth_shmax']=p_dict['checkerboard_shmax_start']+p_dict['checkerboard_gradient']*np.max(np.vstack((cluster_x_grid,cluster_y_grid)),axis=0)
		else:
			raise ValueError('Unknown checkerboard_type ({})'.format(p_dict['checkerboard_type']))

		# Determine synthetic rake
		group_mech_df=mech_df.groupby(by='synth_shmax')
		for synth_shmax_value in list(group_mech_df.groups.keys()):
			tmp_df=group_mech_df.get_group(synth_shmax_value)
			tmp_index=tmp_df.index
			tau=checkerboard.determine_stress_tensor(p_dict['shape_ratio'],synth_shmax_value,0,p_dict['theta'])
			mech_df.loc[tmp_index,'rake']=checkerboard.synth_rake(mech_df.loc[tmp_index,'ddir'].values-90,mech_df.loc[tmp_index,'dip'].values,tau)

	if p_dict['discard_undamped_clusters']:
		discard_cluster_index=cluster_df[~(cluster_df['cluster_id'].isin(damp_df.values.ravel()))].index
		if len(discard_cluster_index):
			discard_cluster_id=cluster_df.loc[discard_cluster_index,'cluster_id'].values
			discard_mech_index=mech_df[mech_df['cluster_id'].isin(discard_cluster_id)].index
			print('Discarding {}/{} ({}%) clusters, representing {}/{} ({}%) of the events, for not having any damped relationships.'.format(
				len(discard_cluster_index),len(cluster_df),np.round(len(discard_cluster_index)/len(cluster_df)*100,1),
				len(discard_mech_index),len(mech_df),np.round(len(discard_mech_index)/len(mech_df)*100,1)))
			cluster_df=cluster_df.drop(discard_cluster_index).reset_index(drop=True)
			mech_df=mech_df.drop(discard_mech_index).reset_index(drop=True)

			le = LabelEncoder()
			le.fit(mech_df['cluster_id'].values)
			mech_df['cluster_id']=le.transform(mech_df['cluster_id'].values)
			damp_df['cluster_id1']=le.transform(damp_df['cluster_id1'].values)
			damp_df['cluster_id2']=le.transform(damp_df['cluster_id2'].values)
			cluster_df['cluster_id']=le.transform(cluster_df['cluster_id'].values)

	return mech_df,cluster_df,damp_df,p_dict


def damp_connect(cluster_df,p_dict):
	'''
	Finds adjacent grid cells, handling non-spatiotemporally uniform results.

	For simple cases (i.e. 2D), this could be done faster. It could also be done faster for complex,
	non-uniform cases with indexing, but the logic may be tougher to follow.
	'''
	print('\tDetermining damping relationships...')
	tic_damping=timer()

	cluster_col=cluster_df.filter(p_dict['dimension_names']).columns.values

	flag_uniform=True
	if 'l' in cluster_col:
		flag_uniform=False

	if flag_uniform:
		# Finds DataFrame of all grid cells contained within domain
		cluster_max=(cluster_df[cluster_col].max().values+1)
		cluster_min=(cluster_df[cluster_col].min().values)
		list_index=[np.arange(min_x,max_x) for min_x,max_x in zip(cluster_min,cluster_max)]
		cluster_ind_dim=[(x.shape[0]) for x in list_index]

		all_cell_coord=np.asarray(list(itertools.product(*list_index)))
		all_cell_coord_df=pd.DataFrame(data=all_cell_coord,columns=cluster_col)

		# Relates the cluster_ids, creating new cluster_ids for empty clusters
		all_cell_coord_df=all_cell_coord_df.merge(cluster_df[np.hstack(('cluster_id',cluster_col))],on=list(cluster_col),how='left').fillna(-1)
		all_cell_coord_df=all_cell_coord_df.astype(int)
		unclustered_index=all_cell_coord_df[all_cell_coord_df['cluster_id']<0].index
		num_unclustered=len(unclustered_index)
		max_cluster_id=all_cell_coord_df['cluster_id'].max()
		new_cluster_id=np.arange(max_cluster_id+1,max_cluster_id+1+num_unclustered)
		all_cell_coord_df.loc[unclustered_index,'cluster_id']=new_cluster_id

		# Applies cluster_ids to the corresponding array location
		cluster_ind=np.zeros(cluster_ind_dim,dtype=int)-1
		## Unpacking operator doesn't work in early python versions. Using tuple for now.
		# cluster_ind[*(all_cell_coord_df[cluster_col].values-all_cell_coord_df[cluster_col].min().values).T]=all_cell_coord_df.index
		cluster_ind[tuple((all_cell_coord_df[cluster_col].values-all_cell_coord_df[cluster_col].min().values).T)]=all_cell_coord_df.index
	else:
		if not({'x','y'}.issubset(cluster_df.columns)):
			raise ValueError('Unable to form nonuniform damping relationships without x and y info')

		l_cluster_df=cluster_df.sort_values(by='l',ascending=False).reset_index(drop=True)
		l_cluster_df['dist']=2**(l_cluster_df['l'])
		l_cluster_df['x']*=l_cluster_df['dist']
		l_cluster_df['y']*=l_cluster_df['dist']
		l_cluster_df['x2']=l_cluster_df['x']+l_cluster_df['dist']
		l_cluster_df['y2']=l_cluster_df['y']+l_cluster_df['dist']

		cluster_col2=['x2','y2']
		if 'z' in cluster_col:
			cluster_col2.append('z')
		if 't' in cluster_col:
			cluster_col2.append('t')
		cluster_ind_dim=l_cluster_df[cluster_col2].max().values+1

		dist_mode=int(np.floor(weighted_mode(l_cluster_df['dist'].values,l_cluster_df['dist'].values)[0][0]))
		if not(dist_mode in l_cluster_df['dist']):
			unique_dist=l_cluster_df['dist'].unique()
			dist_mode=unique_dist[np.argmin(np.abs(unique_dist-dist_mode))]

		# Ensures the x,y dimensions are on the interval dist_mode
		cluster_ind_dim[0]=int(np.ceil(cluster_ind_dim[0]/dist_mode)*dist_mode)
		cluster_ind_dim[1]=int(np.ceil(cluster_ind_dim[1]/dist_mode)*dist_mode)

		# Create the base background grid (cluster_ind)
		cluster_ids_dim=cluster_ind_dim.copy()
		cluster_ids_dim[0]/=dist_mode
		cluster_ids_dim[1]/=dist_mode
		max_cluster_id=l_cluster_df['cluster_id'].max()
		max_base_id=cluster_ids_dim[0]*cluster_ids_dim[1]
		base_background_ids=np.arange(max_cluster_id+1,max_cluster_id+1+max_base_id)
		base_background_ids2d=base_background_ids.reshape((cluster_ids_dim[0],cluster_ids_dim[1]),order='F')
		cluster_ind=np.repeat(base_background_ids2d,dist_mode,1).repeat(dist_mode,0)

		# If 3 or 4 dimensions, expand the background grid, increasing the ind each time
		if len(cluster_col2)==3:
			cluster_ind=np.repeat(cluster_ind[:, :, np.newaxis], cluster_ind_dim[2], axis=2)
			for dim_3 in range(1,cluster_ind_dim[2]):
				cluster_ind[:,:,dim_3]+=(max_base_id*(dim_3))
		elif len(cluster_col2)==4:
			cluster_ind=np.repeat(cluster_ind[:, :, np.newaxis], cluster_ind_dim[2], axis=2)
			cluster_ind=np.repeat(cluster_ind[:, :, :, np.newaxis], cluster_ind_dim[3], axis=3)
			for dim_3 in range(0,cluster_ind_dim[2]):
				for dim_4 in range(0,cluster_ind_dim[3]):
					cluster_ind[:,:,dim_3,dim_4]+=(max_base_id*(dim_3*cluster_ind_dim[3]+dim_4))
		elif len(cluster_col2)>4:
			raise ValueError('Unexpected number of dimensions when determining background grid')

		background_cluster_id=np.arange(max_cluster_id+1,cluster_ind.max()+1)
		background_cell_coord_df=pd.DataFrame(columns=np.hstack(('cluster_id',cluster_col)),index=np.arange(len(background_cluster_id)),dtype=int)
		background_cell_coord_df['cluster_id']=background_cluster_id
		background_cell_coord_df['l']=l_cluster_df.loc[l_cluster_df['dist'].eq(dist_mode).idxmax(),'l']
		background_cell_coord_df['x']=np.tile(np.arange(cluster_ids_dim[0]),np.prod(cluster_ids_dim[1:]))
		background_cell_coord_df['y']=np.tile(np.repeat(np.arange(cluster_ids_dim[1]),cluster_ids_dim[0]),np.prod(cluster_ids_dim[2:]))
		if ('z' in cluster_col) & ('t' in cluster_col):
			background_cell_coord_df['z']=np.tile(np.repeat(np.arange(cluster_ids_dim[2]),cluster_ids_dim[0]*cluster_ids_dim[1]),cluster_ids_dim[3])
			background_cell_coord_df['t']=np.repeat(np.arange(cluster_ids_dim[3]),cluster_ids_dim[0]*cluster_ids_dim[1]*cluster_ids_dim[2])
		elif ('z' in cluster_col):
			background_cell_coord_df['z']=np.repeat(np.arange(cluster_ids_dim[2]),cluster_ids_dim[0]*cluster_ids_dim[1])
		elif ('t' in cluster_col):
			background_cell_coord_df['t']=np.repeat(np.arange(cluster_ids_dim[2]),cluster_ids_dim[0]*cluster_ids_dim[1])

		for cluster_i in range(l_cluster_df.shape[0]):
			if not(('t' in cluster_col) | ('z' in cluster_col)):
				cluster_ind[l_cluster_df.loc[cluster_i,'x']:l_cluster_df.loc[cluster_i,'x2'],
							l_cluster_df.loc[cluster_i,'y']:l_cluster_df.loc[cluster_i,'y2']]=l_cluster_df.loc[cluster_i,'cluster_id']
			elif ('t' in cluster_col) & ('z' in cluster_col):
				cluster_ind[l_cluster_df.loc[cluster_i,'x']:l_cluster_df.loc[cluster_i,'x2'],
							l_cluster_df.loc[cluster_i,'y']:l_cluster_df.loc[cluster_i,'y2'],
							l_cluster_df.loc[cluster_i,'z'],
							l_cluster_df.loc[cluster_i,'t']]=l_cluster_df.loc[cluster_i,'cluster_id']
			elif ('t' in cluster_col):
				cluster_ind[l_cluster_df.loc[cluster_i,'x']:l_cluster_df.loc[cluster_i,'x2'],
							l_cluster_df.loc[cluster_i,'y']:l_cluster_df.loc[cluster_i,'y2'],
							l_cluster_df.loc[cluster_i,'t']]=l_cluster_df.loc[cluster_i,'cluster_id']
			elif ('z' in cluster_col):
				cluster_ind[l_cluster_df.loc[cluster_i,'x']:l_cluster_df.loc[cluster_i,'x2'],
							l_cluster_df.loc[cluster_i,'y']:l_cluster_df.loc[cluster_i,'y2'],
							l_cluster_df.loc[cluster_i,'z']]=l_cluster_df.loc[cluster_i,'cluster_id']
			else:
				raise ValueError('Unable to assign cluster_id to nonuniform grid')

		# Merge the clustered and background clusters together
		all_cell_coord_df=pd.concat([cluster_df,background_cell_coord_df]).reset_index(drop=True)

	# Finds adjacent grid cells
	cluster1_ind=[]
	cluster2_ind=[]
	for dim_i in range(cluster_ind.ndim):
		ind1=np.vstack(np.where(cluster_ind.take(np.arange(1,cluster_ind.shape[dim_i]),axis=dim_i) !=
								cluster_ind.take(np.arange(0,cluster_ind.shape[dim_i]-1),axis=dim_i)))
		ind2=ind1.copy()
		ind2[dim_i]+=1
		## Unpacking operator doesn't work in early python versions. Using tuple for now.
		# cluster1_ind.append(cluster_ind[*ind1])
		# cluster2_ind.append(cluster_ind[*ind2])
		cluster1_ind.append(cluster_ind[tuple(ind1)])
		cluster2_ind.append(cluster_ind[tuple(ind2)])
	cluster1_ind=np.hstack(cluster1_ind)
	cluster2_ind=np.hstack(cluster2_ind)

	damp_df=pd.DataFrame(columns=['cluster_id1','cluster_id2'],index=np.arange(cluster1_ind.shape[0]))
	damp_df['cluster_id1']=all_cell_coord_df.loc[cluster1_ind,'cluster_id'].values
	damp_df['cluster_id2']=all_cell_coord_df.loc[cluster2_ind,'cluster_id'].values

	# Removes any clusters that are not present
	drop_index=all_cell_coord_df[~(all_cell_coord_df['cluster_id'].isin(np.unique(damp_df.values)))].index
	all_cell_coord_df=all_cell_coord_df.drop(drop_index).reset_index(drop=True)

	# Sorts by cluster_id, and puts cluster_id as the first column
	all_cell_coord_df=all_cell_coord_df.sort_values(by='cluster_id').reset_index(drop=True)
	damp_df=damp_df.sort_values(by=['cluster_id1','cluster_id2']).reset_index(drop=True)
	tmp_col=all_cell_coord_df.pop('cluster_id')
	all_cell_coord_df.insert(0, tmp_col.name, tmp_col)

	if len(damp_df):
		print('\t\t{} damped relationships created.'.format(len(damp_df)))
		print('\t\tDamp creation runtime: {0:.2f} sec'.format(timer()-tic_damping))
		return all_cell_coord_df,damp_df
	else:
		return cluster_df,damp_df


def rotate_points(origin, points, angle):
	'''
	Rotates points clockwise by a given angle around a given origin.
	'''
	angle_r=np.radians(-angle)
	cos_angle_r=np.cos(angle_r)
	sin_angle_r=np.sin(angle_r)

	rotated_points=np.zeros(points.shape)
	rotated_points[:,0]=origin[0]+cos_angle_r*(points[:,0]-origin[0])-sin_angle_r*(points[:,1]-origin[1])
	rotated_points[:,1]=origin[1]+sin_angle_r*(points[:,0]-origin[0])+cos_angle_r*(points[:,1]-origin[1])

	return rotated_points


'''
Default parameters used for pySATSI.

Author: Robert J. Skoumal (rskoumal@usgs.gov)
'''

import numpy as np
import pandas as pd

p_dict={
	# ------- File paths -------
	'control_file':'', # Path to control file
	'event_file':'', # Path to the mechanism file
	'damp_file':'', # Path to pre-determined damp relationships
	'binloc_file':'', # Path to pre-determined bin locations
	'project_dir':'', # The directory the output files will be saved

	# ------- Study area -------
	'min_lon':-np.inf, # Min longitude, deg
	'max_lon':np.inf, # Max longitude, deg
	'min_lat':-np.inf, # Min latitude, deg
	'max_lat':np.inf, # Max latitude, deg
	'min_x':-np.inf, # Min horizontal x location, kilometers
	'min_y':np.inf, # Min horizontal y location, kilometers
	'max_x':-np.inf, # Min horizontal x location, kilometers
	'max_y':np.inf, # Min horizontal y location, kilometers
	'min_depth':-np.inf, # Min vertical z location, kilometers
	'max_depth':np.inf, # Min vertical z location, kilometers
	'base_lat':np.nan, # Reference latitude for coordinate conversion
	'earliest_event_time':pd.Timestamp(''), # Earliest event time considered
	'latest_event_time':pd.Timestamp(''), # Earliest event time considered

	# ------- Clustering -------
	'minev':15, # Minimum number of mechanisms in a cell to be considered
	'cluster_type':'', # 'grid' or 'quadroot'
	'hspace':np.inf, # Grid cell size used when creating equally spaced grids, km
	'xspace':[], # Grid cell size for horizontal (x) grids
	'yspace':[], # Grid cell size for horizontal (y) grids
	'hspace_min':-np.inf, # Minimum horizontal grid size when producing quadroot clusters, km
	'hspace_max':np.inf, # Maximum horizontal grid size when producing quadroot clusters, km
	'hspace_values':[], # Rather than providing $hspace_min & $hspace_max, you can provide the grid size values directly, km
	'depth_breaks':[], # Depth intervals (km)
	'time_breaks':[], # List of times (strings) that will divide the study time. Example format: ['1990-04-01 01:02:03','2024-01-01 20:10:05']
	'rotate_deg':0, # Degrees counterclockwise to rotate horizontal grid cells

	# ------- Damping -------
	'do_damping':True,
	'write_damp_file':False, # If True, the file containing the damping relationships (damp_outfile) will be created
	'do_iterative':False, # If True, damp values will be iteratively sampled as the optimal knee is located. If False, every damp value is considered.
	'damp_iteration_max':50, # When iteratively sampling damping parameters, this sets the maximum number of iterations.
	'damp_smoothing':True, # If True, damping is 'smoothed' by considering cells without events
	'discard_undamped_clusters':False, # If True, clusters that are not damped with another cluster are discarded.
	'diag_buffer_flag':False, # If True, cells that are diagonal from one another are damped
	'damp_param':-1, # If >0, this damping parameter will be used instead of using the model length vs data variance tradeoff curve
	'damp_values':[0,0.02,0.05,0.1,0.2,0.3,0.4,0.6,0.8,1,1.2,1.4,1.6,1.8,2,2.25,2.5,2.75,3,3.5,4,5,6,7,8,9,10,15,20,50],# Damping parameters considered
	'num_additional_damp':0, # Additional number of measurements surrounding the knee e value
	'report_empty_clusters':False, # If True, damped estimates for clusters without mechanisms are reported

	# ------- Bootstrapping -------
	'nboot':30, # Number of bootstap resamplings for uncertainty estimates (>1000 recommended)
	'fracplane':50, # Percent chance that the nodal plane is the fault plane (50 if chosen randomly).
	'confidence':95, # Confidence level (e.g., 95 for 95%)
	'randomize_events':True, # If True, mechanisms will be randomly selected for each bootstrap iteration
	'randomize_nodalplane':True, # If True, the nodal plane that represents the fault plane will be randomly selected for each bootstrap iteration
	'mech_uncertainty':0, # The median of the normally distributed errors applied to the nodal planes.
	'num_cpus':1, # Number of cores to run the boostrapping in parallel. Set to 0 to use all available, or 1 to run in serial.

	# ------- Fault plane instability -------
	'niterations':0, # Number of instability iterations
	'iteration_weight':0.5, # Amount to weigh the fault probability for each iteration (0.5-1.0). 0.5 is equal weight, 1.0 doesn't consider previous weight.
	'instability_weight':0.5, # Amount to weigh relative instability calculation (0.0-1.0).
	'final_instability':False, # On the final iteration, an instability_weight and iteration_weight of 1.0 is used.
	'convergence_threshold':1.0, # Stops the iterations when the mean change in fault probability % is less than the convergence_threshold
	'iterative_damping':False, # If True, recomputes the damping parameter following an iteration.
	'friction':[0.4,0.45,0.5,0.55,0.6,0.65,0.7,0.75,0.8,0.85,0.9,0.95,1.0], # Friction values to consider

	# ------- Mechanism selection -------
	'max_fault_plane_uncertainty':np.nan, # Maximum fault plane uncertainty of mechs to consider.
	'max_polarity_misfit':np.nan, # Maximum polarity misfit of mechs to consider.
	'min_prob_mech':np.nan, # Minimum mechanism probably to consider.
	'min_sta_distribution_ratio':np.nan, # Minimum station distribution ratio of the mechanisms to consider.

	# ------- Checkerboard -------
	'checkerboard_type':'', # 'traditional' or 'gradient'. If empty, no checkerboard test is done.
	'shape_ratio':0.5,
	'theta':0, # dip of sigma1 for the checkerboard (0: horizontal, 90: vertical)
	'checkerboard_hspace':10, # length of checkerboard grid, km
	# -- Traditional checkerboard --
	'checkerboard_shmax1':0, # Shmax azimuth of one of the traditional checkerboard values, degrees
	'checkerboard_shmax2':45, # Shmax azimuth of the other traditional checkerboard value, degrees
	# -- Gradient checkerboard --
	'checkerboard_gradient':1, # rotation of shmax clockwise per gradient checkerboard_hspace, degrees
	'checkerboard_shmax_start':5, # Shmax value in the SW corner

	# ------- Output variables and measurement precision -------
	'coordinate_precision':6, # number of decimal places to round lon/lats
	'angle_precision':1, # number of decimal places to round measurements/solutions
	'vector_precision':6, # number of decimal places to round vector & phi results
	'merge_bin_info':True, # If True, the stress file will contain bin location/time information that is reported in $binloc_outfile

	# ------- Plotting -------
	'plots':[], # Can include 'stereonet', 'shmax', 'shmax_error', 'phi', 'aphi','damp_tradeoff'

	# ------- Cosmetic -------
	'print_runtime':True,

	# ------- Additional (hidden) variables -------
	# -- Filenames produced by pySATSI --
	'event_outfile':'', # Output path to the binned mechanism file
	'stress_outfile':'', # Output path to stress results
	'rawstress_outfile':'', # Output path to the raw boostrap stress results
	'binloc_outfile':'', # Output path to the bin locations
	'tradeoff_outfile':'', # Output path to the damp misfit results
	'damp_outfile':'', # Output path to the damping relationships. Created if write_damp_file==True
	# -- Cosmetic variables. Only used for printing version information --
	'version_string':'v0.2',
	'version_date':'2025-05-20',
	# -- Variables used internally. Don't modify these unless you know what you're doing. :)
	'dimension_names':['x','y','z','l','t'], # Allowed dimension names
	'min_grid_spacing':np.nan,
	'do_horizontal_gridding':False,
	'do_vertical_gridding':False,
	'do_temporal_gridding':False,
}


'''
Functions used for bootstrapping and uncertainty computations.

Author: Robert J. Skoumal (rskoumal@usgs.gov)
'''

# Standard libraries
import multiprocessing

# External libraries
import pandas as pd
import numpy as np
import scipy.stats as st
rng=np.random.default_rng(123) # Used to produce reproducable bootstrapped results.

# Local libraries
import functions.stress as stress
import functions.fun as fun
import functions.knee as knee

def compute_boot(mech_df,damp_df,p_dict):
	'''
	nboot: Number of bootstrap iterations to perform.
		For nboot==1 or the first boostrap iteration, it will compute the stresses using the provided nodal planes and events.
		For nboot>1, it will randomize according to the $randomize_events and $randomize_nodalplane options.
	fracplane: Chance the nodal plane is the fault plane. Either a float or an array of floats
	'''

	num_iterations=p_dict['niterations']
	if num_iterations<1:
		num_iterations=1

	for iter_x in range(num_iterations):
		boot_stress_df=[]
		if p_dict['num_cpus']==1: # Runs bootstrapping in serial
			for boot_x in range(p_dict['nboot']):
				select_mech_df=mech_df.copy()
				if boot_x!=0:
					select_mech_df=randomize_mech(select_mech_df,p_dict)
				boot_stress_df.append(stress.slickenside(select_mech_df,damp_df,p_dict['damp_param'],p_dict,boot_x=boot_x,iter_x=iter_x))
		else: # Runs bootstrapping in parallel
			print('\tComputing bootstraps in parallel...')
			pool=multiprocessing.Pool(processes=p_dict['num_cpus'])
			async_results=[]
			for boot_x in range(p_dict['nboot']):
				select_mech_df=mech_df.copy()
				if boot_x!=0:
					select_mech_df=randomize_mech(select_mech_df,p_dict)
				async_results.append(pool.apply_async(stress.slickenside,
							(select_mech_df,damp_df,p_dict['damp_param'],p_dict),dict(boot_x=boot_x,iter_x=iter_x)))
			pool.close()
			pool.join()
			for result in async_results:
				boot_stress_df.append(result.get())

		boot_stress_df=pd.concat(boot_stress_df).reset_index(drop=True)

		# Use fault instability to inform fault plane
		if p_dict['niterations']>0:
			if (iter_x>=(num_iterations-2)) and (p_dict['final_instability']==True):
				instability_weight=1.
				iteration_weight=1.
				damp_df=pd.DataFrame(columns=['cluster_id1','cluster_id2'])
			else:
				instability_weight=p_dict['instability_weight']
				iteration_weight=p_dict['iteration_weight']

			mech_df['instability'],mech_df['aux_instability']=fault_instability(boot_stress_df,mech_df,p_dict)
			Idiff=mech_df['instability'].values**2-mech_df['aux_instability'].values**2
			abs_Idiff=np.abs(Idiff)

			# Determine the fault probability
			old_fault_probability=mech_df['fault_probability'].values
			mech_df['fault_probability']=100*(iteration_weight*
											(((Idiff/(instability_weight*abs_Idiff-instability_weight+1)+1)/2)-old_fault_probability/100) +
											old_fault_probability/100)

			# Checks for convergence
			mean_prob_diff=np.mean(np.abs(mech_df['fault_probability'].values-old_fault_probability))
			print('\t\tMean change of fault probability: {}%'.format(mean_prob_diff.round(2)))
			if mean_prob_diff<p_dict['convergence_threshold']:
				if iter_x<(num_iterations-1):
					print('\tReached the convergence threshold of {}%. Stopping iterations.'.format(p_dict['convergence_threshold']))
					break
			if p_dict['iterative_damping']:
				if iter_x<(num_iterations-1):
					iteration_damp_misfit_df,p_dict['damp_param']=knee.find_damp(mech_df,damp_df,p_dict)
	if p_dict['iterative_damping']:
		return boot_stress_df,iteration_damp_misfit_df
	else:
		return boot_stress_df,None

def randomize_mech(mech_df,p_dict):
	'''
	Randomizes the nodal plane for the mechanisms

	Input:
		mech_df: DataFrame of [ddir, ddir, dip, dip2, rake1, rake2, cluster_id]
		fracplane: Percentage that the nodal plane is the fault plane. Either a float or an array of floats
		randomize_events: Whether the selection of events are randomized, boolean.
		randomize_nodalplane: Whether the selection of nodal planes are randomized, boolean.
	Output:
		select_mech_df: DataFrame containing randomized nodal planes
	'''
	random_mech_df=mech_df.copy()
	nobs=len(random_mech_df)
	if p_dict['randomize_events']: # Randomizes which events are selected
		rand1=rng.choice(nobs,nobs)
		random_mech_df=random_mech_df.copy().loc[rand1,:].reset_index(drop=True)

	# Applies noise and randomizes which nodal plane is the fault plane
	if random_mech_df['fault_plane_uncertainty'].any():
		random_mech_df=mech_noise(random_mech_df)
	elif p_dict['randomize_nodalplane']:
		random_mech_df['ddir2'],random_mech_df['dip2'],random_mech_df['rake2']=fun.aux_plane_ddir(random_mech_df['ddir'].values,random_mech_df['dip'].values,random_mech_df['rake'].values)
		rand2=rng.uniform(size=nobs)*100
		if 'fault_probability' in random_mech_df:
			ind_rand2=np.where(rand2>random_mech_df['fault_probability'])[0]
		else:
			ind_rand2=np.where(rand2>p_dict['fracplane'])[0]
		random_mech_df.loc[ind_rand2,['ddir','dip','rake']]=random_mech_df.loc[ind_rand2,['ddir2','dip2','rake2']].values
		random_mech_df=random_mech_df.drop(columns=['ddir2','dip2','rake2'])

	return random_mech_df


def avguncert_azimuth_plunge(azimuth,plunge,round=1,confidence_thresh=95.):
	'''
	Given azimuths and plunges (in degrees), computes the eigenvectors of the
	covariance matrix and returns the mean azimuth in plunge (in degrees).
	'''

	# Converts plunge,azimuth (degrees) to lon,lat (radians)
	lat=np.radians(90-plunge)
	lon=0
	theta=np.radians(azimuth)

	# Converts lon,lat to xyz
	x=np.cos(lat)*np.cos(lon)
	y=np.cos(lat)*np.sin(lon)
	z=np.sin(lat)

	# Rotate lon,lat coords by theta degrees
	y_2=y*np.cos(theta)+z*np.sin(theta)
	z_2=-y*np.sin(theta)+z*np.cos(theta)
	r=np.sqrt(x**2+y_2**2+z_2**2)
	lat=np.arcsin(z_2/r)
	lon=np.arctan2(y_2,x)

	# Computes the antipodes
	x=np.cos(lat)*np.cos(lon)
	y=np.cos(lat)*np.sin(lon)
	z=np.sin(lat)
	r=np.sqrt(x**2+y**2+z**2)
	lat_antipode=np.arcsin(-z/r)
	lon_antipode=np.arctan2(-y,-x)

	# Concats the antipodes to the lon,lat arrays
	lon,lat=np.hstack([lon,lon_antipode]),np.hstack([lat,lat_antipode])
	xyz=np.zeros((len(lat),3))
	xyz[:,0]=np.cos(lat)*np.cos(lon)
	xyz[:,1]=np.cos(lat)*np.sin(lon)
	xyz[:,2]=np.sin(lat)

	# Computes eigenvectors
	cov=np.cov(xyz.T)
	eigvalues,eigvectors=np.linalg.eigh(cov)
	sort_index=eigvalues.argsort()
	vecs=eigvectors[:,sort_index]
	x,y,z=vecs[:,-1]

	# Calculate azimuth and plunge (in radians)
	azimuth_rads=np.arctan2(z,y)
	r=np.sqrt(x**2+y**2+z**2)
	if r==0:
		plunge_rads=np.arcsin(1)
	else:
		plunge_rads=np.arcsin(x/r)

	# Convert azimuth and plunge to degrees
	mean_plunge=np.degrees(plunge_rads)
	mean_azimuth=np.degrees(azimuth_rads)
	mean_azimuth=90-mean_azimuth
	if mean_azimuth<0:
		mean_azimuth+=360

	# If upwards plunge, get antipode
	if mean_plunge<0:
		mean_plunge*=-1
		mean_azimuth-=180
		if mean_azimuth<0:
			mean_azimuth+=360

	mean_azimuth=np.round(mean_azimuth,round)
	mean_plunge=np.round(mean_plunge,round)

	# Calculate confidence bounds
	n_measurements=len(azimuth)
	j=int(np.floor((n_measurements-1)*((100.-confidence_thresh)/100.)))
	dot_conf=np.dot(xyz[:int(xyz.shape[0]/2)],np.asarray([x,y,z]))
	ind_conf=np.argsort(np.abs(dot_conf))[j]
	conf_azimuth=azimuth[ind_conf]
	conf_plunge=plunge[ind_conf]

	if dot_conf[ind_conf]<0:
		conf_azimuth-=180

	uncert_azimuth=fun.az_diff_180(mean_azimuth,conf_azimuth)
	uncert_plunge=np.abs(mean_plunge-conf_plunge)

	# Round uncertainties
	uncert_azimuth=np.round(uncert_azimuth,round)
	uncert_plunge=np.round(uncert_plunge,round)

	return mean_azimuth,mean_plunge,uncert_azimuth,uncert_plunge

def uncert(boot_stress_df,p_dict):
	'''
	Calculates the average and variability range of the bootstrapped results
	'''
	# Group bootstrapped results by cluster_id
	boot_stress_df=boot_stress_df.sort_values(by=['cluster_id','boot'])

	# If no bootstrapping was done, no uncertainties need to be calculated
	if boot_stress_df['boot'].max()==0:
		boot_uncert_df=boot_stress_df[['cluster_id','phi','az1','pl1','az2','pl2','az3','pl3',]]
	else: # Calculates uncertainties
		group_boot_stress_df=boot_stress_df.groupby(by='cluster_id')
		r_keys=list(group_boot_stress_df.groups.keys())

		# Create empty uncert dataframe
		boot_uncert_df=pd.DataFrame(index=np.arange(len(r_keys)),
								columns=['cluster_id','phi','uncert_phi',
											'az1','az1_uncert','pl1','pl1_uncert',
											'az2','az2_uncert','pl2','pl2_uncert',
											'az3','az3_uncert','pl3','pl3_uncert'],
											dtype=float)
		boot_uncert_df['cluster_id']=r_keys
		boot_uncert_df['cluster_id']=boot_uncert_df['cluster_id'].astype(int)

		# Loop over each cluster
		for r_x,r_key in enumerate(r_keys):
			stress_df=group_boot_stress_df.get_group(r_key)

			# If there's only one record, no need to calculate uncertainties
			if len(stress_df)==1:
				boot_uncert_df.loc[r_x,['phi','az1','pl1','az2','pl2','az3','pl3']]=stress_df.iloc[0][['phi','az1','pl1','az2','pl2','az3','pl3']]
				continue

			# Computes mean azimuth/plunge using the eigenvectors of the covariance matrix
			boot_uncert_df.loc[r_x,['az1','pl1','az1_uncert','pl1_uncert']]=avguncert_azimuth_plunge(
					stress_df['az1'].values,stress_df['pl1'].values,
					round=p_dict['angle_precision'],confidence_thresh=p_dict['confidence'])
			boot_uncert_df.loc[r_x,['az2','pl2','az2_uncert','pl2_uncert']]=avguncert_azimuth_plunge(
					stress_df['az2'].values,stress_df['pl2'].values,
					round=p_dict['angle_precision'],confidence_thresh=p_dict['confidence'])
			boot_uncert_df.loc[r_x,['az3','pl3','az3_uncert','pl3_uncert']]=avguncert_azimuth_plunge(
					stress_df['az3'].values,stress_df['pl3'].values,
					round=p_dict['angle_precision'],confidence_thresh=p_dict['confidence'])

			# Computes phi stats
			boot_uncert_df.loc[r_x,'phi']=np.round(stress_df['phi'].mean(),p_dict['vector_precision'])
			if len(stress_df['phi'].unique())>1:
				tmp_interval=st.t.interval(p_dict['confidence']/100,stress_df.shape[0]-1,loc=boot_uncert_df.loc[r_x,['phi']],scale=st.sem(stress_df['phi'].values))
				boot_uncert_df.loc[r_x,'uncert_phi']=np.round(np.max(np.abs(boot_uncert_df.loc[r_x,'phi']-tmp_interval)),p_dict['vector_precision'])
			else:
				boot_uncert_df.loc[r_x,'uncert_phi']=0.

	return boot_uncert_df

def mech_noise(select_mech_df):
	'''
	Adds random noise to the mechanism solutions
	'''
	noise_mech_df=select_mech_df.copy()
	strike_r=np.radians(noise_mech_df['ddir'].values-90)
	dip_r=np.radians(noise_mech_df['dip'].values)
	rake_r=np.radians(noise_mech_df['rake'].values)

	num_mechs=len(strike_r)

	# Fault normals
	n=np.zeros((num_mechs,3))
	n[:,0]=-np.sin(dip_r)*np.sin(strike_r)
	n[:,1]=np.sin(dip_r)*np.cos(strike_r)
	n[:,2]=-np.cos(dip_r)

	# Slip direction
	u=np.zeros((num_mechs,3))
	u[:,0]=np.cos(rake_r)*np.cos(strike_r) + np.cos(dip_r)*np.sin(rake_r)*np.sin(strike_r)
	u[:,1]=np.cos(rake_r)*np.sin(strike_r) - np.cos(dip_r)*np.sin(rake_r)*np.cos(strike_r)
	u[:,2]=-np.sin(rake_r)*np.sin(dip_r)

	rand2=rng.uniform(size=num_mechs)*100
	if 'fault_probability' in noise_mech_df:
		tmp_bool=rand2<noise_mech_df['fault_probability'].values
	else:
		tmp_bool=rand2<50
	i0_ind=np.where(tmp_bool)[0]
	i1_ind=np.where(~tmp_bool)[0]

	b=np.zeros((len(strike_r),3))
	b[i0_ind]=np.cross(n[i0_ind],u[i0_ind])
	b[i1_ind]=np.cross(n[i1_ind],u[i1_ind])

	fault_plane_uncertainty=noise_mech_df['fault_plane_uncertainty'].values
	if 'aux_plane_uncertainty' in noise_mech_df.columns:
		fault_plane_uncertainty[i1_ind]=noise_mech_df['aux_plane_uncertainty'].values[i1_ind]

	# Random fault normal
	n_deviation_random=np.radians(fault_plane_uncertainty*np.random.randn(num_mechs))
	b_deviation_random=np.radians(fault_plane_uncertainty*np.random.randn(num_mechs))

	# Random noisy azimuth
	n_azimuth_random=np.pi*2*np.random.random(num_mechs)
	b_azimuth_random=np.pi*2*np.random.random(num_mechs)

	# Applies random noise to n
	n_perpendicular=u*np.sin(n_azimuth_random)[:,np.newaxis]+b*np.cos(n_azimuth_random)[:,np.newaxis]
	n_noise_component=n_perpendicular*np.sin(n_deviation_random)[:,np.newaxis]
	n_noisy=n+n_noise_component
	n_noisy=n_noisy/np.linalg.norm(n_noisy,2,axis=1)[:,np.newaxis]

	# Applies random noise to b
	b_perpendicular=n_noisy*np.sin(b_azimuth_random)[:,np.newaxis]+u*np.cos(b_azimuth_random)[:,np.newaxis]
	b_noise_component=b_perpendicular*np.sin(b_deviation_random)[:,np.newaxis]
	b_noisy=b+b_noise_component
	b_noisy=b_noisy/np.linalg.norm(b_noisy,2,axis=1)[:,np.newaxis]

	# Determines noisy slip direction
	u_noisy=np.cross(n_noisy,b_noisy)
	u_noisy*=np.sign(np.sum(u*u_noisy,axis=1))[:,np.newaxis]

	n=n_noisy.copy()
	u=u_noisy.copy()
	n[i1_ind]=u_noisy[i1_ind]
	u[i1_ind]=n_noisy[i1_ind]

	n1=n.copy()
	u1=u.copy()
	n2=u.copy()
	u2=n.copy()

	# Ensures vertical components are negative
	pos_n1_ind=np.where(n1[:,2]>0)[0]
	n1[pos_n1_ind]*=-1
	u1[pos_n1_ind]*=-1
	pos_n2_ind=np.where(n2[:,2]>0)[0]
	n2[pos_n2_ind]*=-1
	u2[pos_n2_ind]*=-1

	# Determine strike, dip, rake of noisy mech
	r_dip=np.arccos(-n1[:,2])
	r_strike=np.arcsin(-n1[:,0]/np.sqrt(n1[:,0]**2+n1[:,1]**2))
	dip=np.degrees(r_dip)
	strike=np.degrees(r_strike)
	rake=np.degrees(np.arcsin(-u1[:,2]/np.sin(r_dip)))

	# Determining quadrants
	n1_quad_ind=np.where(n1[:,1]<0)[0]
	strike[n1_quad_ind]=180-strike[n1_quad_ind]
	cos_rake=u1[:,0]*np.cos(np.radians(strike))+u1[:,1]*np.sin(np.radians(strike))
	u1_quad_ind=np.where(cos_rake<0)[0]
	rake[u1_quad_ind]=180-rake[u1_quad_ind]

	strike[strike<0]+=360
	rake[rake<-180]+=360
	rake[rake>180]-=360

	noise_mech_df['ddir']=(strike+90)%360
	noise_mech_df['dip']=dip
	noise_mech_df['rake']=rake

	return noise_mech_df

def fault_instability(boot_stress_df,mech_df,p_dict):
	'''
	Calculates the fault instability for the primary and auxiliary
	nodal planes following the instability parameter of Vavryčuk (2014).
	Takes the median instability considering the provided friction values.
	'''
	group_mech_df=mech_df.groupby(by='cluster_id')
	group_boot_stress_df=boot_stress_df.groupby(by='cluster_id')
	instability=np.zeros(mech_df.shape[0]);instability[:]=np.nan
	aux_instability=np.zeros(mech_df.shape[0]);aux_instability[:]=np.nan

	for cluster_id in list(group_mech_df.groups.keys()):
		cluster_mech_df=group_mech_df.get_group(cluster_id)
		cluster_boot_stress_df=group_boot_stress_df.get_group(cluster_id)

		ddir=cluster_mech_df['ddir'].values
		dip=cluster_mech_df['dip'].values
		rake=cluster_mech_df['rake'].values

		# Determines aux nodal plane
		aux_ddir,aux_dip,aux_rake=fun.aux_plane_ddir(ddir,dip,rake)

		# Converts ddir,dip to radians
		z=np.radians(ddir)
		dip[dip==90]=89.99999
		z2=np.radians(dip)

		# Converts aux ddir,dip to radians
		aux_z=np.radians(aux_ddir)
		aux_ddir[aux_ddir==90]=89.99999
		aux_z2=np.radians(aux_dip)

		# Normal vector to plane
		normal=np.zeros((len(z),3))
		normal[:,0]=np.sin(z)*np.sin(z2)
		normal[:,1]=np.cos(z)*np.sin(z2)
		normal[:,2]=np.cos(z2)

		# Normal vector to aux plane
		aux_normal=np.zeros((len(aux_z),3))
		aux_normal[:,0]=np.sin(aux_z)*np.sin(aux_z2)
		aux_normal[:,1]=np.cos(aux_z)*np.sin(aux_z2)
		aux_normal[:,2]=np.cos(aux_z2)

		# Determines stress tensor
		tau=np.zeros((len(cluster_boot_stress_df),3,3))
		tau[:,0,0]=cluster_boot_stress_df.loc[:,'st00']
		tau[:,0,1]=cluster_boot_stress_df.loc[:,'st01']
		tau[:,0,2]=cluster_boot_stress_df.loc[:,'st02']
		tau[:,1,1]=cluster_boot_stress_df.loc[:,'st11']
		tau[:,1,2]=cluster_boot_stress_df.loc[:,'st12']
		tau[:,2,2]=cluster_boot_stress_df.loc[:,'st22']
		tau[:,1,0]=tau[:,0,1]
		tau[:,2,0]=tau[:,0,2]
		tau[:,2,1]=tau[:,1,2]
		tau=np.sum(tau,axis=0)
		tau=tau/np.linalg.norm(tau,2)
		tau_rot=tau.copy()
		tau_rot[0,0]=tau[1,1]
		tau_rot[1,1]=tau[0,0]
		tau_rot[0,2]=tau[1,2]*-1
		tau_rot[1,2]=tau[0,2]*-1
		tau_rot[1,0]=tau_rot[0,1]
		tau_rot[2,0]=tau_rot[0,2]
		tau_rot[2,1]=tau_rot[1,2]
		tau=tau_rot.copy()

		lam,vecs=np.linalg.eig(tau)

		# Sort eigenvalues
		sort_ind=lam.argsort()
		lam=lam[sort_ind]
		vecs=vecs[:,sort_ind]

		# Shape ratio
		shape_ratio=(lam[0]-lam[1])/(lam[0]-lam[2])

		normal=normal[:,[1,0,2]]
		normal[:,2]*=-1
		aux_normal=aux_normal[:,[1,0,2]]
		aux_normal[:,2]*=-1

		# Fault and aux plane normals * principal stresses
		n=np.sum(normal[:,:,np.newaxis]*vecs,axis=1)
		aux_n=np.sum(aux_normal[:,:,np.newaxis]*vecs,axis=1)

		# Shear and normal traction along plane (Vavryčuk EQ 17,18)
		tau_shear=np.sqrt(n[:,0]**2+(1-2*shape_ratio)**2*n[:,1]**2.+n[:,2]**2-(n[:,0]**2+(1-2*shape_ratio)*n[:,1]**2-n[:,2]**2)**2)
		tau_normal=(n[:,0]**2+(1-2*shape_ratio)*n[:,1]**2-n[:,2]**2)

		# Shear and normal traction along aux plane (Vavryčuk EQ 17,18)
		aux_tau_shear=np.sqrt(aux_n[:,0]**2+(1-2*shape_ratio)**2*aux_n[:,1]**2.+aux_n[:,2]**2-(aux_n[:,0]**2+(1-2*shape_ratio)*aux_n[:,1]**2-aux_n[:,2]**2)**2)
		aux_tau_normal=(aux_n[:,0]**2+(1-2*shape_ratio)*aux_n[:,1]**2-aux_n[:,2]**2)

		# Computes instability for plane (Vavryčuk EQ 16)
		cluster_instability=(tau_shear[:,np.newaxis]-p_dict['friction']*(tau_normal[:,np.newaxis]-1))/(p_dict['friction']+np.sqrt(1+p_dict['friction']**2))
		cluster_instability=np.median(cluster_instability,axis=1)

		# Computes instability for aux plane (Vavryčuk EQ 16)
		cluster_aux_instability=(aux_tau_shear[:,np.newaxis]-p_dict['friction']*(aux_tau_normal[:,np.newaxis]-1))/(p_dict['friction']+np.sqrt(1+p_dict['friction']**2))
		cluster_aux_instability=np.median(cluster_aux_instability,axis=1)

		instability[cluster_mech_df.index]=cluster_instability
		aux_instability[cluster_mech_df.index]=cluster_aux_instability

	return instability, aux_instability


'''
Functions for identifying the knee of the tradeoff curve.

Author: Robert J. Skoumal (rskoumal@usgs.gov)
'''

# Standard libraries
import os

# External libraries
import numpy as np
import pandas as pd
from scipy.signal import argrelextrema

# Local libraries
import functions.stress as stress
import functions.fun as fun
# import functions.plotting as plotting


def find_damp(in_mech_df,damp_df,p_dict):
	'''
	Evaluates damp values and uses the variance vs length tradeoff curve to determine the optimal damp value.
	'''
	mech_df=in_mech_df.copy()
	if 'fault_probability' in mech_df.columns:
		if (mech_df['fault_probability']<50).any():
			use_aux_index=mech_df.loc[mech_df['fault_probability']<50].index
			mech_df.loc[use_aux_index,['ddir','dip','rake']]=np.vstack(fun.aux_plane_ddir(mech_df.loc[use_aux_index,'ddir'].values,mech_df.loc[use_aux_index,'dip'].values,mech_df.loc[use_aux_index,'rake'].values)).T

	G,D,slick=stress.slickenside(mech_df,damp_df,-1,p_dict,case='slick')

	# Computing damp misfit using precomputed G, D, slick
	damp_misfit_df=pd.DataFrame(index=np.arange(len(p_dict['damp_values'])),columns=['e','variance','length'],dtype=float)

	if p_dict['do_iterative']:
		num_damps=len(p_dict['damp_values'])
		damp_int=np.unique([0,int(np.floor(num_damps/3)),int(np.floor(num_damps/2)),int(np.floor(num_damps/3)*2),len(p_dict['damp_values'])-1])
		if len(damp_int)!=5:
			print('Unable to iteratively determine damping parameters. Considering all values.')
			p_dict['do_iterative']=False
		damp_misfit_df.loc[damp_int,'e']=p_dict['damp_values'][damp_int]

	if p_dict['do_iterative']: # Finds the knee using steps of damp values
		damp_int_all=damp_int
		break_next=False
		for R in range(p_dict['damp_iteration_max']):
			print('\tIteration {}'.format(R))
			for damp_x,damp_value in enumerate(p_dict['damp_values'][damp_int]):
				print('\t\t{}/{}: e={}'.format(damp_x,len(damp_int)-1,damp_value))
				damp_misfit_df.loc[damp_int[damp_x],['variance','length']]=stress.slickenside(mech_df,damp_df,damp_value,p_dict,G=G,D=D,slick=slick,case='misfit')
			damp_misfit_df.loc[damp_int,'e']=p_dict['damp_values'][damp_int]

			# Finds the knee
			knee_index=find_knee(damp_misfit_df['length'].dropna().values, damp_misfit_df['variance'].dropna().values)
			if break_next:
				break

			damp_int_all=damp_misfit_df['length'].dropna().index.values
			new_damp_int=[int(np.floor((damp_int_all[knee_index-1]+damp_int_all[knee_index])/2)),
						int(np.ceil((damp_int_all[knee_index]+damp_int_all[knee_index+1])/2))]
			if (new_damp_int[0]==(damp_int_all[knee_index]-1)) & (new_damp_int[1]==(damp_int_all[knee_index]+1)):
				break_next=True
			damp_int=new_damp_int

			if R==(p_dict['damp_iteration_max']-1):
				print('\tMaximum number of iterations reached.')

		damp_misfit_df=damp_misfit_df.dropna().reset_index(drop=True)

	else: # Finds the knee using all damp values
		damp_misfit_df['e']=p_dict['damp_values']
		for damp_x,damp_value in enumerate(p_dict['damp_values']):
			print('\t{}/{}: e={}'.format(damp_x,len(p_dict['damp_values'])-1,damp_value))
			damp_misfit_df.loc[damp_x,['variance','length']]=stress.slickenside(mech_df,damp_df,damp_value,p_dict,G=G,D=D,slick=slick,case='misfit')

	# Corrects any non-increasing lengths
	inc_index=damp_misfit_df[damp_misfit_df['length'].diff()>0].index
	if len(inc_index):
		trial_x=0
		while trial_x<10:
			damp_misfit_df.loc[inc_index-1,'length']=damp_misfit_df.loc[inc_index,'length'].values
			inc_index=damp_misfit_df[damp_misfit_df['length'].diff()>0].index
			if len(inc_index)==0:
				break
			trial_x+=1

	# Finds the knee
	knee_index=find_knee(damp_misfit_df['length'].values, damp_misfit_df['variance'].values)

	if knee_index is None:
		raise ValueError('Unable to find the optimal damp value. Increasing the number and range of damp_values may solve the issue.')
	damp_param=damp_misfit_df.loc[knee_index,'e']

	# Looks at damping parameters surrounding the knee
	if p_dict['num_additional_damp']>0:
		# Ensures there are samples with indicies above & below the knee
		if (knee_index<=(damp_misfit_df.shape[0]-2)) & (knee_index>=1):
			additional_damp_values=np.linspace(damp_misfit_df.loc[knee_index-1,'e'],damp_misfit_df.loc[knee_index+1,'e'],p_dict['num_additional_damp']+2)[1:-1]
			additional_damp_values=np.unique(np.round(additional_damp_values,5))

			# Finds the variance and length for the additional damping parameters
			add_damp_misfit_df=pd.DataFrame(index=np.arange(len(additional_damp_values)),columns=['e','variance','length'],dtype=float)
			add_damp_misfit_df['e']=additional_damp_values
			for damp_x,damp_value in enumerate(additional_damp_values):
				print('\t(additional) {}/{}: e={}'.format(damp_x,len(additional_damp_values)-1,damp_value))
				add_damp_misfit_df.loc[damp_x,['variance','length']]=stress.slickenside(mech_df,damp_df,damp_value,p_dict,G=G,D=D,slick=slick,case='misfit')
			damp_misfit_df=pd.concat([damp_misfit_df,add_damp_misfit_df]).sort_values(by=['e']).reset_index(drop=True)

			# Finds the knee
			knee_index=find_knee(damp_misfit_df['length'].values, damp_misfit_df['variance'].values)
			if knee_index is None:
				raise print('Unable to find the knee with the additional values. Using original damp parameter of e={}'.format(damp_param))
			damp_param=damp_misfit_df.loc[knee_index,'e']

	damp_misfit_df['knee_flag']=False
	damp_misfit_df.loc[knee_index,'knee_flag']=True
	print('\tSelected damp parameter of e={}'.format(damp_param))

	damp_misfit_df.round(6).to_csv(p_dict['tradeoff_outfile'],float_format='%f',index=False)

	return damp_misfit_df,damp_param

def find_knee(x_values,y_values):
	'''
	Given a curve of x_values and y_values, returns the index of the first knee found
	'''
	# Normalize values
	x_normalized=(x_values-np.min(x_values))/(np.max(x_values)-np.min(x_values))
	y_normalized=(y_values-np.min(y_values))/(np.max(y_values)-np.min(y_values))

	# Calculate the diff curve
	y_normalized=y_normalized.max()-y_normalized

	# Normalize diff curve
	y_difference=y_normalized-x_normalized
	x_difference=x_normalized.copy()

	# Find the local max
	maxima_indices=argrelextrema(y_difference, np.greater_equal)[0]
	y_difference_maxima=y_difference[maxima_indices]

	# Find the local min
	minima_indices=argrelextrema(y_difference, np.less_equal)[0]

	# Calculate thresholds
	Tmx=y_difference_maxima-(np.abs(np.diff(x_normalized).mean()))

	# If no max, no knee can be found
	if not maxima_indices.size:
		return None

	max_threshold_index=0
	knee=None

	# Iterate through the diff curve
	for x,_ in enumerate(x_difference):
		if x < maxima_indices[0]:
			continue

		# End of the curve and no knee found
		if x==(len(x_difference)-1):
			break

		# If at a local max, increment the max threshold index
		if (maxima_indices==x).any():
			threshold=Tmx[max_threshold_index]
			threshold_index=x
			max_threshold_index+=1
		# Now for local min
		if (minima_indices==x).any():
			threshold=0.0

		if y_difference[x+1]<threshold:
			knee=x_values[threshold_index]
			break

	# Return knee index
	if knee:
		return threshold_index
	else:
		return None




'''
Functions used to read/write files for pySATSI.

Author: Robert J. Skoumal (rskoumal@usgs.gov)
'''

# Standard libraries
import os
import multiprocessing
import argparse
import importlib.util
from timeit import default_timer as timer

# External libraries
import numpy as np
import pandas as pd

def read_command_line(p_dict):
	'''
	Reads user inputs provided on the command line.
	'''
	parser = argparse.ArgumentParser()
	parser.add_argument('control_file',nargs='?')
	for p_dict_var in p_dict.keys():
		parser.add_argument('--'+p_dict_var)
	args = vars(parser.parse_args())
	for p_dict_var in p_dict.keys():
		if args[p_dict_var] is not None:
			dtype=type(p_dict[p_dict_var])
			if dtype==bool:
				if args[p_dict_var].lower() in ['true','1']:
					p_dict[p_dict_var]=True
				elif args[p_dict_var].lower() in ['false','0']:
					p_dict[p_dict_var]=False
				else:
					raise ValueError(('Expected a boolean for the command-line declared variable \'{}\''+\
						'The provided value ({}) is not a boolean.').format(p_dict_var,args[p_dict_var]))
			else:
				p_dict[p_dict_var]=dtype(args[p_dict_var])
	return p_dict

def read_control_file(p_dict):
	'''
	Reads the user's control file.
	If the control file does not contain a variable, the default value in p_dict is used.
	If the control file contains an unexpected variable, it is ignored after printing a warning message.
	This file should follow the format:
		$var_name
		var_value
		...
	Where "$var_name" is the name of the variable with a "$" as the first non-whitespace character,
	and "var_value" is the value of the variable on the following line.

	User text on all other lines will be ignored, provided the first non-whitespace character on the line is not a "$".
	User text will also be ignored following a space (" ") after "$var_name"
	'''
	if not(os.path.exists(p_dict['control_file'])):
		raise ValueError('Control file ({}) does not exist.'.format(p_dict['control_file']))

	# Reads lines from the control file
	with open(p_dict['control_file']) as f:
		lines=f.read().splitlines()
		lines=[line.strip() for line in lines]

	# Removes any text following a '#'
	for line_x in range(len(lines)):
		if '#' in lines[line_x]:
			lines[line_x]=lines[line_x][:lines[line_x].find('#')].strip()

	# Identifies the variable names and their line numbers
	var_name=[]
	var_line=[]
	value_line=[]
	for line_x,line in enumerate(lines):
		if line: # If not empty
			if line[0]=='$':
				tmp_var_name=line.split()[0][1:]
				tmp_var_name=tmp_var_name.split('#')[0] # In case the user didn't put a space after the variable name.
				if tmp_var_name in p_dict.keys():
					var_name.append(tmp_var_name)
					var_line.append(line_x)
				else:
					raise ValueError('Unknown variable name ({}) in the control file on line {}:\n\t{}'.format(tmp_var_name,line_x,line))
			elif line[0]!='#':
				value_line.append(line_x)

	# Ensures there are not duplicate variables
	var_set=set()
	duplicate_var=[x for x in var_name if x in var_set or var_set.add(x)]
	if duplicate_var:
		raise ValueError('Duplicate variable name in control file: \"{}\"'.format(duplicate_var[0]))

	# Ensures that every value line corresponds with a variable
	corr_bool=np.isin( (np.asarray(value_line)-1),var_line)
	if not(np.all(corr_bool)):
		prob_ind=np.where(corr_bool==False)[0][0]
		raise ValueError('Unable to associate a value (\"{}\") on line {} with a variable in the control file. If this is a variable, start the line with a \'$\'. If this is intended to be ignored, start the line with a \'#\'.'.format(lines[value_line[prob_ind]],value_line[prob_ind]))

	if (var_line[-1]+1)==len(lines):
		raise ValueError('Unable to associate a value with the final variable in the control file.')

	# Determines the variable values that follow the variable name lines
	var_value=[]
	for line_x in var_line:
		tmp_var_value=lines[line_x+1].strip()

		if len(tmp_var_value)==0:
			raise ValueError('No value is provided for the variable on line {} in the control file.'.format(line_x+1))
		if tmp_var_value[0]=='#':
			raise ValueError('Value (\"{}\") on line {} in the control file starts with a \'#\'.'.format(tmp_var_value,line_x+1) )
		if tmp_var_value.isdigit(): # if int
			tmp_var_value=int(tmp_var_value)
		elif tmp_var_value.replace('.','',1).lstrip('-').isdigit(): # if float
			tmp_var_value=float(tmp_var_value)
		elif (tmp_var_value=='True'):
			tmp_var_value=True
		elif (tmp_var_value=='False'): # if bool
			tmp_var_value=False
		elif (',' in tmp_var_value): # list divided by commas
			if '[' in tmp_var_value:
				tmp_var_value=tmp_var_value.replace('[','',1)
			if ']' in tmp_var_value:
				tmp_var_value=tmp_var_value.replace(']','',1)
			tmp_var_value=tmp_var_value.split(',')
			tmp_var_value=[s.strip() for s in tmp_var_value] # Removes any whitespace
		elif (('\'' in tmp_var_value) or ('\"' in tmp_var_value)): # If variables are divided by quotation marks (' or ")
			tmp_var_value=tmp_var_value.replace('\"','\'')
			print(tmp_var_value)
			quote_ind=[i for i, letter in enumerate(tmp_var_value) if letter == '\'']
			if not((len(quote_ind) % 2)==0):
				raise ValueError('Unable to parse the variable line due to an odd number of quotations:\n\t{}'.format(tmp_var_value))
			tmp=[]
			for tmp_x in range(0,len(quote_ind),2):
				tmp.append(tmp_var_value[quote_ind[tmp_x]:quote_ind[tmp_x+1]])
			tmp_var_value=tmp
		var_value.append(tmp_var_value)

	# Assigns the variable values to the dictionary
	for name,value in zip(var_name,var_value):
		p_dict[name]=value

	# Some basic variable type fixing
	p_dict['hspace_values']=np.atleast_1d(p_dict['hspace_values']).astype(float)
	p_dict['damp_values']=np.atleast_1d(p_dict['damp_values']).astype(float)
	p_dict['xspace']=np.atleast_1d(p_dict['xspace']).astype(float)
	p_dict['yspace']=np.atleast_1d(p_dict['yspace']).astype(float)
	p_dict['depth_breaks']=np.atleast_1d(p_dict['depth_breaks']).astype(float)
	if type(p_dict['time_breaks'])!=list:
		p_dict['time_breaks']=list([p_dict['time_breaks']])
	p_dict['time_breaks']=[pd.Timestamp(t) for t in p_dict['time_breaks']]
	p_dict['plots']=np.atleast_1d(p_dict['plots'])

	return p_dict


def qc_params(p_dict):
	'''
	Does various quality control checks on the user parameters
	'''
	# Creates event filepaths
	if not(p_dict['event_outfile']):
		p_dict['event_outfile']=os.path.join(p_dict['project_dir'],'bin_mechs.csv') # binned mechanisms
	if not(p_dict['binloc_outfile']):
		p_dict['binloc_outfile']=os.path.join(p_dict['project_dir'],'bin_locations.csv') # locations of bins
	if not(p_dict['damp_outfile']):
		p_dict['damp_outfile']=os.path.join(p_dict['project_dir'],'damp_pairs.csv') # pairs of neighboring bins for damping
	if not(p_dict['tradeoff_outfile']):
		p_dict['tradeoff_outfile']=os.path.join(p_dict['project_dir'],'damp_tradeoff.csv')

	# Standard outfiles
	if not(p_dict['rawstress_outfile']):
		p_dict['rawstress_outfile']=os.path.join(p_dict['project_dir'],'raw_stress.csv')
	if not(p_dict['stress_outfile']):
		p_dict['stress_outfile']=os.path.join(p_dict['project_dir'],'stress.csv')

	# Requires inputs to be provided
	if not(p_dict['project_dir']):
		raise ValueError('A project directory ($project_dir) is required.')
	if not(p_dict['event_file']):
		raise ValueError('An event file ($event_file) is required.')

	# Ensures the input files exist
	if len(p_dict['event_file']):
		if not(os.path.exists(p_dict['event_file'])):
			raise ValueError('Event input file (event_file) not found: {}'.format(p_dict['event_file']))
	if len(p_dict['damp_file']):
		if not(os.path.exists(p_dict['damp_file'])):
			raise ValueError('Damp input file (damp_file) not found: {}'.format(p_dict['damp_file']))
	if len(p_dict['binloc_file']):
		if not(os.path.exists(p_dict['binloc_file'])):
			raise ValueError('Bin location file (binloc_file) not found: {}'.format(p_dict['binloc_file']))

	# Ensure the cluster file exists if the damp file exists
	if os.path.isfile(p_dict['binloc_file']):
		if not(os.path.isfile(p_dict['damp_outfile'])):
			raise ValueError('When providing a damp file ({}), the bin locations ({}) should also be provided'.format(p_dict['damp_outfile'],p_dict['binloc_file']))

	# Creates output folder(s), if needed
	for outfile in [p_dict['event_outfile'],p_dict['binloc_outfile'],p_dict['damp_outfile']]:
		if os.path.dirname(outfile):
			os.makedirs(os.path.dirname(outfile),exist_ok=True)

	# Ensures damp values are floats
	p_dict['damp_values']=np.asarray(p_dict['damp_values']).astype(float)

	# Ensures damp values are sorted
	if (len(np.unique(p_dict['damp_values']))!=len(p_dict['damp_values'])):
		raise ValueError('Damping values should be unique, but duplicates were found: {}'.format(p_dict['damp_values']))

	# There must be at least one boot
	if p_dict['nboot']<1:
		raise ValueError('There must be at least one iteration. Set nboot>=1.')

	# Determines the number of cores to use
	if p_dict['num_cpus']!=1:
		num_cpus_avail=multiprocessing.cpu_count()-1
		if (p_dict['num_cpus']==0): # Uses all available cores
			p_dict['num_cpus']=num_cpus_avail
		elif (p_dict['num_cpus']>num_cpus_avail): # If the num requested > available cores, use all available
			print('You requested {} cores, but only {} are available. Using all available cores instead.'.format(p_dict['num_cpus'],num_cpus_avail))
			p_dict['num_cpus']=num_cpus_avail

	# If there are five or fewer damp values, raise an error
	if p_dict['do_damping']:
		if len(p_dict['damp_values'])<5:
			raise ValueError('There should be at least 5 $damp_values, but there are currently {} value(s) provided: {}'.format(len(p_dict['damp_values']),p_dict['damp_values']))

	# Ensures depth breaks are increasing
	if len(p_dict['depth_breaks']):
		p_dict['depth_breaks']=np.sort(p_dict['depth_breaks'])

	if (p_dict['rotate_deg']<0) | (p_dict['rotate_deg']>90):
		raise ValueError('Grid rotations ($rotate_deg) should be 0-90°, but the provided value is: {}'.format(p_dict['rotate_deg']))

	if len(p_dict['hspace_values']):
		# If directly providing hspace values, ensures the other horizontal spacing variables are not provided
		for space_name in ['hspace_min','hspace_max','hspace']:
			print(p_dict[space_name])
			if np.isfinite(p_dict[space_name]):
				raise ValueError('When providing horizontal bin sizes ($hspace_values), the horizontal bin size variable (${}) is not necessary.'.format(space_name))
		for space_name in ['xspace','yspace']:
			if len(p_dict[space_name]):
				raise ValueError('When providing horizontal bin sizes ($hspace_values), the horizontal bin size variable (${}) is not necessary.'.format(space_name))

		# Ensures hspace values are a factor of 2 (allowing for skips)
		p_dict['hspace_values']=np.sort(p_dict['hspace_values'])
		if not(np.all((np.log2(p_dict['hspace_values']/p_dict['hspace_values'][0]) % 1)==0)):
			raise ValueError('$hspace_values must increase by a factor of 2')
		p_dict['xspace']=p_dict['hspace_values']
		p_dict['yspace']=p_dict['hspace_values']
		p_dict['hspace_values']=[]
	elif np.isfinite(p_dict['hspace']):
		# If providing a uniform hspace value, ensures the other horizontal spacing variables are not provided
		for space_name in ['hspace_min','hspace_max']:
			if np.isfinite(p_dict[space_name]):
				raise ValueError('When providing a uniform horizontal bin size ($hspace), the horizontal bin size variable (${}) is not necessary.'.format(space_name))
		for space_name in ['xspace','yspace']:
			if len(p_dict[space_name]):
				raise ValueError('When providing horizontal bin sizes ($hspace_values), the horizontal bin size variable (${}) is not necessary.'.format(space_name))

		p_dict['xspace']=np.asarray([p_dict['hspace']])
		p_dict['yspace']=np.asarray([p_dict['hspace']])
		p_dict['hspace']=np.inf
	elif np.isfinite(p_dict['hspace_min']) | np.isfinite(p_dict['hspace_max']):
		if not(np.isfinite(p_dict['hspace_min'])):
			raise ValueError('When providing a minimum horizontal bin size ($hspace_min), the maximum horizontal bin size ($hspace_max) must be provided.'.format(space_name))
		if not(np.isfinite(p_dict['hspace_max'])):
			raise ValueError('When providing a maximum horizontal bin size ($hspace_max), the minimum horizontal bin size ($hspace_min) must be provided.'.format(space_name))
		if (p_dict['hspace_min']<=0) & (p_dict['hspace_max']<=0):
			raise ValueError('Both the minimum ($hspace_min) and maximum ($hspace_max) horizontal bin sizes must be > 0')
		num_spacings=np.round((np.log2(p_dict['hspace_max'])-np.log2(p_dict['hspace_min']))+1,10)
		if (num_spacings).is_integer():
			num_spacings=int(num_spacings)
		else:
			potential_hspace_max=p_dict['hspace_min']*2**(int(np.round(num_spacings))-1)
			raise ValueError('hspace_max ({} km) and hspace_min ({} km) must be related by log2 (e.g., hspace_max={})'.format(p_dict['hspace_max'],p_dict['hspace_min'],potential_hspace_max))
		p_dict['xspace']=p_dict['hspace_min']*2**(np.arange(num_spacings))
		p_dict['yspace']=p_dict['xspace']

	# If multiple xspace and yspace values are being considered, they must have the same values
	if (len(p_dict['xspace'])>1) | (len(p_dict['yspace'])>1):
		if len(p_dict['xspace'])!=len(p_dict['yspace']):
			raise ValueError('$xspace ({}) and $yspace ({}) must be the same length. If you are considering multiple horizontal spacings, consider using the $hspace variable.'.format(p_dict['xspace'],p_dict['yspace']))
		# if len(p_dict['xspace'])>1:
		if not((p_dict['xspace']==p_dict['yspace']).all()):
			raise ValueError('When using multiple xspace ({}) and yspace ({}) values, they must be identical'.format(p_dict['xspace'],p_dict['yspace']))

	# Creates convinence flags to determine if spatial/temporal gridding is needed
	if len(p_dict['xspace']) | len(p_dict['yspace']):
		p_dict['do_horizontal_gridding']=True
	if len(p_dict['depth_breaks']):
		p_dict['do_vertical_gridding']=True
	if len(p_dict['time_breaks'])>0:
		p_dict['do_temporal_gridding']=True

	if p_dict['damp_smoothing']:
		if p_dict['discard_undamped_clusters']:
			print('Because $damp_smoothing==True, all clusters will be damped. Setting $discard_undamped_clusters==False')
			p_dict['discard_undamped_clusters']=False

	# Checks if the cluster type agrees with the expected data types
	if p_dict['cluster_type']=='grid':
		if len(p_dict['hspace_values'])>1:
			raise ValueError('When doing a \'grid\' cluster_type, only one grid spacing can be considered.')
	elif p_dict['cluster_type']=='quadroot':
		if len(p_dict['hspace_values'])<2:
			raise ValueError('Quadroot clustering expects multiple hspace values.')

	# Determine minimum grid interval
	if len(p_dict['checkerboard_type']):
		p_dict['min_grid_spacing']=np.min([p_dict['checkerboard_hspace'],p_dict['hspace_min']])
	else:
		p_dict['min_grid_spacing']=np.min(p_dict['hspace_min'])

	# Ensures friction values are an array of floats
	p_dict['friction']=np.atleast_1d(np.asarray(p_dict['friction'])).astype(float)

	# Ensures plot types are lowercase
	p_dict['plots']=[plot_type.lower() for plot_type in p_dict['plots']]
	# Only plots stereonets if the spacing is the same
	if 'stereonet' in p_dict['plots']:
		if np.isfinite(p_dict['hspace_min']):
			if p_dict['hspace_min']!=p_dict['hspace_max']:
				print('WARNING: Cannot plot stereonets when doing non-unfiform spatial grids.')
				p_dict['plots']=p_dict['plots'][p_dict['plots']!='stereonet']
	# Ensures the plots are a recognized type
	for plot_type in p_dict['plots']:
		if not(plot_type in ['stereonet','shmax','shmax_error','aphi','phi','damp_tradeoff']):
			raise ValueError('Unknown plot type: {}'.format(plot_type))

	if len(p_dict['plots']):
		if (spec := importlib.util.find_spec('matplotlib')) is None:
			raise ImportError('Unable to find the matplotlib module. Either install matplotlib (pip install \'pySATSI[plot]\') or remove the following plot requests: {}'.format(p_dict['plots']))

	return p_dict


def read_mech_file(p_dict):
	'''
	Reads in file containing earthquake focal mechanism

	Must always contain the columns:
		'dip' and 'rake'
	Must always contain either:
		'strike' or 'ddir'
	If considering horizontal event location, must contain the corresponding columns:
		'lat' and/or 'lon' (*or* 'x_km' and/or 'y_km')
	If considering vertical event location, must contain:
		'depth'
	If considering event timing, must contain:
		'time'
	If using pre-determined fault plane probability, must contain:
		'fault_probability'
	If using individual fault plane uncertainties, must contain:
		'fault_plane_uncertainty'
	If using individual aux plane uncertainties, must contain:
		'aux_plane_uncertainty'
	If using pre-clustered mechanisms, must contain:
		'cluster_id' (*or* some combination of 'x', 'y', 'z', 't', 'l')
	If selecting mechanisms based on qualities, must contain at least one of the following:
		'fault_plane_uncertainty', 'aux_plane_uncertainty', 'polarity_misfit', 'prob_mech', 'sta_distribution_ratio'
	'''
	mech_df=pd.read_csv(p_dict['event_file'])

	if 'latitude' in mech_df.columns:
		mech_df=mech_df.rename(columns={'latitude':'lat'})
	elif 'event_lat' in mech_df.columns:
		mech_df=mech_df.rename(columns={'event_lat':'lat'})
	if 'longitude' in mech_df.columns:
		mech_df=mech_df.rename(columns={'longitude':'lon'})
	elif 'event_lon' in mech_df.columns:
		mech_df=mech_df.rename(columns={'event_lon':'lon'})
	if 'event_depth' in mech_df.columns:
		mech_df=mech_df.rename(columns={'event_depth':'depth'})
	elif 'origin_depth_km' in mech_df.columns:
		mech_df=mech_df.rename(columns={'origin_depth_km':'depth'})

	if np.sum(mech_df.columns.isin(['lat']))>1:
		raise ValueError('Multiple latitude columns provided in event_file ({})'.format(p_dict['event_file']))
	if np.sum(mech_df.columns.isin(['lon']))>1:
		raise ValueError('Multiple longitude columns provided in event_file ({})'.format(p_dict['event_file']))
	if np.sum(mech_df.columns.isin(['depth']))>1:
		raise ValueError('Multiple depth columns provided in event_file ({})'.format(p_dict['event_file']))

	if not({'dip','rake'}.issubset(mech_df.columns)):
		raise ValueError('The binned mechanism file (event_outfile) must contain the columns: [cluster_id, dip, rake]')
	if not('ddir' in mech_df.columns):
		if not('strike' in mech_df.columns):
			raise ValueError('The binned mechanism file (event_outfile) must contain either the column [strike] or [ddir]')

	if (not(p_dict['do_horizontal_gridding'])) & (not(p_dict['do_vertical_gridding'])) & (not(p_dict['do_temporal_gridding'])):
		if (not(mech_df.columns.isin(p_dict['dimension_names']).any())):
			p_dict['do_damping']=False

	if (p_dict['mech_uncertainty']>0) and ('fault_plane_uncertainty' in mech_df.columns):
		raise ValueError('When individual fault plane uncertainties (\'fault_plane_uncertainty\' column) are present in the mechanism file, the constant $mech_uncertainty should be 0.')
	if ('aux_plane_uncertainty' in mech_df.columns) and (not('fault_plane_uncertainty' in mech_df.columns)):
		raise ValueError('When the \'aux_plane_uncertainty\' column is present in the mechanism file, the \'fault_plane_uncertainty\' column should be present too.')
	if (p_dict['mech_uncertainty']>0):
		mech_df['fault_plane_uncertainty']=p_dict['mech_uncertainty']
	if not('fault_plane_uncertainty' in	mech_df.columns):
		mech_df['fault_plane_uncertainty']=0

	if not(p_dict['randomize_nodalplane']):
		mech_df['fault_probability']=100
	if not('fault_probability' in mech_df.columns):
		mech_df['fault_probability']=p_dict['fracplane']

	# # If dip direction present, convert to strike
	# if 'ddir' in mech_df.columns:
	# 	mech_df['strike']=mech_df['ddir']-90
	# mech_df['strike']%=360

	# If strike direction present, convert to dip direction
	if 'strike' in mech_df.columns:
		mech_df['ddir']=mech_df['strike']+90
	mech_df['ddir']%=360

	# Ensures ddir, dip, and rake are floats
	mech_df[['ddir','dip','rake']]=mech_df[['ddir','dip','rake']].astype(float)

	# If dip is 0, sets to a small number instead
	mech_df.loc[mech_df['dip']==0,'dip']=0.00001

	if 'x' in mech_df.columns:
		mech_df['x']=mech_df['x'].astype(int)
	if 'y' in mech_df.columns:
		mech_df['y']=mech_df['y'].astype(int)
	if 'z' in mech_df.columns:
		mech_df['z']=mech_df['z'].astype(int)
	if 't' in mech_df.columns:
		mech_df['t']=mech_df['t'].astype(int)

	if 'fault_probability' in mech_df.columns:
		mech_df['fault_probability']=mech_df['fault_probability'].astype(float)

	# Ensures cluster_id is an integer
	if 'cluster_id' in mech_df.columns:
		mech_df['cluster_id']=mech_df['cluster_id'].astype(int)

	# Removes any earthquakes that fall outside of the selected study area
	drop_flag=np.zeros(len(mech_df),dtype=bool)
	if np.isfinite(p_dict['min_lon']) or np.isfinite(p_dict['max_lon']):
		if not('lon' in mech_df.columns):
			raise ValueError('When considering a min or max longitude ($min_lon, $max_lon), there must be a \'lon\' column in the mechanism file ({})'.
					format(p_dict['event_file']))
		drop_flag+=(mech_df['lon'].values<p_dict['min_lon'])+(mech_df['lon'].values>p_dict['max_lon'])
	if np.isfinite(p_dict['min_lat']) or np.isfinite(p_dict['max_lat']):
		if not('lon' in mech_df.columns):
			raise ValueError('When considering a min or max latitude ($min_lat, $max_lat), there must be a \'lat\' column in the mechanism file ({})'.
					format(p_dict['event_file']))
		drop_flag+=(mech_df['lat'].values<p_dict['min_lat'])+(mech_df['lat'].values>p_dict['max_lat'])
	if np.isfinite(p_dict['min_x']) or np.isfinite(p_dict['max_x']):
		if not('x_km' in mech_df.columns):
			raise ValueError('When considering a min or max horizontal x location ($min_x, $max_x), there must be a \'x_km\' column in the mechanism file ({})'.
					format(p_dict['event_file']))
		drop_flag+=(mech_df['x_km'].values<p_dict['min_x'])+(mech_df['x_km'].values>p_dict['max_x'])
	if np.isfinite(p_dict['min_y']) or np.isfinite(p_dict['max_y']):
		if not('y_km' in mech_df.columns):
			raise ValueError('When considering a min or max horizontal y location ($min_y, $max_y), there must be a \'y_km\' column in the mechanism file ({})'.
					format(p_dict['event_file']))
		drop_flag+=(mech_df['y_km'].values<p_dict['min_y'])+(mech_df['y_km'].values>p_dict['max_y'])
	if np.isfinite(p_dict['min_depth']) or np.isfinite(p_dict['max_depth']):
		if not('y_km' in mech_df.columns):
			raise ValueError('When considering a min or max depth ($min_depth, $max_depth), there must be a \'depth\' column in the mechanism file ({})'.
					format(p_dict['event_file']))
		drop_flag+=(mech_df['y_km'].values<p_dict['min_y'])+(mech_df['y_km'].values>p_dict['max_y'])

	if np.any(drop_flag):
		print('\tIgnoring {}/{} ({}%) mechanisms for being outside of the selected area'.
			format(np.sum(drop_flag),len(mech_df),np.round(np.sum(drop_flag)/len(mech_df)*100,1)))
		mech_df=mech_df.drop(mech_df.loc[drop_flag].index).reset_index(drop=True)

	# Drops focal mechanisms that don't meet the specified quality criteria
	if (~np.isnan([p_dict['max_fault_plane_uncertainty'],p_dict['max_polarity_misfit'],
			p_dict['min_prob_mech'],p_dict['min_sta_distribution_ratio']])).any():

		drop_flag=np.zeros(mech_df.shape[0],dtype=bool)
		if ~np.isnan(p_dict['max_fault_plane_uncertainty']):
			if not('fault_plane_uncertainty' in mech_df.columns):
				raise ValueError(('When setting a max fault plane uncertainty (max_fault_plane_uncertainty={}), there should be a \'fault_plane_uncertainty\' '+
					 '(and optionally an \'auxillary_plane_uncertainty\') column in the mechanism file.').format(p_dict['max_fault_plane_uncertainty']))
			# If aux plane uncertainties also provided, compute average fault plane uncertainty
			if 'aux_plane_uncertainty' in mech_df.columns:
				drop_flag[((mech_df['fault_plane_uncertainty']+mech_df['aux_plane_uncertainty'])/2)>p_dict['max_fault_plane_uncertainty']]=True
			else: # Otherwise, use just the fault plane uncertainty
				drop_flag[mech_df['fault_plane_uncertainty']>p_dict['max_fault_plane_uncertainty']]=True

		if ~np.isnan(p_dict['max_polarity_misfit']):
			if not('polarity_misfit' in mech_df.columns):
				raise ValueError(('When setting a max polarity misfit (max_polarity_misfit={}), there should be a \'polarity_misfit\' '+
					 'column in the mechanism file.').format(p_dict['max_polarity_misfit']))
			drop_flag[mech_df['polarity_misfit']>p_dict['max_polarity_misfit']]=True

		if ~np.isnan(p_dict['min_prob_mech']):
			if not('prob_mech' in mech_df.columns):
				raise ValueError(('When setting a min mechanism probability (prob_mech={}), there should be a \'prob_mech\' '+
					 'column in the mechanism file.').format(p_dict['min_prob_mech']))
			drop_flag[mech_df['prob_mech']<p_dict['min_prob_mech']]=True

		if ~np.isnan(p_dict['min_sta_distribution_ratio']):
			if not('sta_distribution_ratio' in mech_df.columns):
				raise ValueError(('When setting a min station distribution ratio (min_sta_distribution_ratio={}), there should be a \'sta_distribution_ratio\' '+
					 'column in the mechanism file.').format(p_dict['min_sta_distribution_ratio']))
			drop_flag[mech_df['sta_distribution_ratio']<p_dict['min_sta_distribution_ratio']]=True

		if drop_flag.any():
			print('\tIgnoring {}/{} ({}%) mechanisms for not meeting the mechanism quality criteria.'.
				format(np.sum(drop_flag),len(mech_df),np.round(np.sum(drop_flag)/len(mech_df)*100,1)))

	if p_dict['do_vertical_gridding']:
		# Ensures depth column exists if considering event depths
		if not('depth' in mech_df.columns):
			raise ValueError('When considering event depths, mechanisms must include a \'depth\' column.')

	if 'time' in mech_df.columns:
		mech_df['time']=pd.to_datetime(mech_df['time'])

	# If considering timing, the 'time' column must be present
	if len(p_dict['time_breaks']) or not(pd.isnull(p_dict['earliest_event_time'])) or not(pd.isnull(p_dict['latest_event_time'])):
		if not('time' in mech_df.columns):
			raise ValueError('When considering event times (with time_breaks, earliest_event_time, or latest_event_time), mechanisms must include a \'time\' column.')

		drop_index=mech_df[(mech_df['time']<p_dict['earliest_event_time']) | (mech_df['time']>p_dict['latest_event_time'])].index
		if len(drop_index):
			print('\tIgnoring {}/{} ({}%) mechanisms for falling outside of the selected time window ({} - {})'.format(\
				len(drop_index),len(mech_df),np.round(len(drop_index)/len(mech_df)*100,1),p_dict['earliest_event_time'],p_dict['latest_event_time']))
			mech_df=mech_df.drop(drop_index).reset_index(drop=True)

	# Selects only the columns of interest
	mech_df=mech_df.filter(['ddir','dip','rake','fault_probability','fault_plane_uncertainty', 'aux_plane_uncertainty',
						 	'lat','lon','x_km','y_km','depth','time','x', 'y', 'z', 't', 'l','cluster_id'])

	if len(mech_df)==0:
		print('There are no mechanisms in selected dataset. Quitting.')
		quit()

	return mech_df,p_dict


def read_damp_file(p_dict):
	'''
	Reads a .csv file containing damping relationships.

	Must contain the columns:
		'cluster_id1', 'cluster_id2'
	'''
	damp_df=pd.read_csv(p_dict['damp_file'])

	# If the cluster_id1 and cluster_id2 columns already exist, return those columns
	if set(['cluster_id1','cluster_id2']).issubset(damp_df.columns):
		damp_df=damp_df[['cluster_id1','cluster_id2']].astype(int)
		return damp_df

	# If individual dimension names are given, read those in
	found_dim_s=[]
	for dim_s in p_dict['dimension_names']:
		if (dim_s+'1') in damp_df.columns:
			if (dim_s+'2') in damp_df.columns:
				found_dim_s.append(dim_s)
			else:
				raise ValueError('In damp_file ({}), dimension {}1 was found but {}2 was missing.'.format(p_dict['damp_file'],dim_s,dim_s))
	if len(found_dim_s):
		use_dim_col=[]
		for dim_s in found_dim_s:
			use_dim_col.append(dim_s+'1')
		for dim_s in found_dim_s:
			use_dim_col.append(dim_s+'2')
		damp_df=damp_df[use_dim_col].astype(int)
		return damp_df

	raise ValueError(('In damp_file ({}), the following column names (with a \'1\' or \'2\' suffix) are acceptable but were not found:'+
					'\n\t\'cluster_id\'\n\t\t- OR -\n\t{}').format(p_dict['damp_file'],p_dict['dimension_names']))

def read_cluster_file(p_dict):
	'''
	Reads file of predetermined cluster locations
	'''
	if len(p_dict['binloc_file']):
		if os.path.exists(p_dict['binloc_file']):
			cluster_df=pd.read_csv(p_dict['binloc_file'])
			return cluster_df
	raise ValueError('The binloc file ({}) is empty'.format(p_dict['binloc_file']))

def write_misfit_file(mech_df,p_dict):
	'''
	Writes file containing the mechanism misfits
	'''
	misfit_mech_df=mech_df.copy()
	misfit_mech_df['strike']=(misfit_mech_df['ddir']-90)%360
	misfit_mech_df[['strike','dip','rake','misfit_angle']]=misfit_mech_df[['strike','dip','rake','misfit_angle']].round(p_dict['angle_precision'])
	if 'fault_probability' in misfit_mech_df.columns:
		misfit_mech_df['fault_probability']=misfit_mech_df['fault_probability'].round(p_dict['angle_precision']) # Using angle precision for %
	misfit_mech_df[['tau_mag']]=misfit_mech_df[['tau_mag']].round(p_dict['vector_precision'])
	misfit_mech_df=misfit_mech_df.filter(np.hstack([['strike','dip','rake','cluster_id'],p_dict['dimension_names'],['misfit_angle','tau_mag','fault_probability']]))
	misfit_mech_df.to_csv(p_dict['event_outfile'],index=False)
	return 0